#' Data Upload Server Component
#' 
#' This server module handles the server-side logic for data upload,
#' including reading the uploaded file, and storing it in a reactive value.
#'
#' @param id Unique identifier for the server module, ensuring isolated scope
#'           within the Shiny application.
#'
#' @return
#' @export
#'
#' @examples
#' data <- DataUploadServer("DataUploadModuleUI")
DataUploadServer <- function(id) {
  moduleServer(id, function(input, output, session) {
    
    ns <- session$ns
    
    # Store data files
    data <- reactiveVal(NULL)
    
    # Upload Data files
    observeEvent(input$uploadData, {
      
      if(input$fileoption == "Processed Data from CogentAP") {
        
        if (is.null(input$file1)) {
          shinyalert('Missing Data', 'Please upload the data.', type = "error", confirmButtonCol = "#1c82e0")
          return(NULL)
        } 
        
        req(input$file1)
        
        withProgress(message = paste("Uploading Data..."), value = 0, {
          
          # Load CCN files
          incProgress(1/6, detail = "Reading uploaded file...")
          ccn_data_files <-readRDS(input$file1$datapath)
          
          incProgress(2/6, detail = "Processing raw data...")
          ginkgo_data_m <-  as.matrix(ccn_data_files@assays[["ginkgo_data"]]$counts) %>%
            as.data.frame() %>%
            pivot_longer(cols = everything(), names_to = "variable", values_to = "value")
          cl_names <- gsub("_trimmed$", "", ginkgo_data_m$variable)
          ginkgo_data_m$sample <- gsub("_[^_]*$", "", cl_names)
          
          incProgress(3/6, detail = "Normalizing data...")
          ginkgo_data_normalized <- sweep(as.matrix(ccn_data_files@assays[["ginkgo_data"]]$counts) + 1, 2, colMeans(ccn_data_files@assays[["ginkgo_data"]]$counts + 1), '/')
          ginkgo_data_normalized <- cbind(ccn_data_files[["RNA"]][["GC_Content"]], ginkgo_data_normalized)
          ginkgo_data_normalized_m <- as.matrix(ginkgo_data_normalized) %>%
            as.data.frame() %>%
            pivot_longer(cols = -GC_Content, names_to = "variable", values_to = "value")
          cl_names <- gsub("_trimmed$", "", ginkgo_data_normalized_m$variable)
          
          ginkgo_data_normalized_m$sample <- gsub("_[^_]*$", "", cl_names)
          incProgress(4/6, detail = "Calculating Gini scores...")
          gini_scores <- apply(ccn_data_files@assays[["ginkgo_data"]]$counts, MARGIN=2, FUN=Gini)
          cl_names <- gsub("_trimmed$", "", names(gini_scores))
          sample <- gsub("_[^_]*$", "", cl_names)
          gini_scores_df <- data.frame(sample=sample, gini=gini_scores)
          
          # Update data
          incProgress(5/6, detail = "Finalizing data upload...")
          data(list(ccn_data_files = ccn_data_files, 
                    gini_scores = gini_scores_df, 
                    normalized_data_melted = ginkgo_data_normalized_m, 
                    ginkgo_data_melted = ginkgo_data_m, 
                    normalized_data = ginkgo_data_normalized))
          
        })
        
        shinyalert("Data Uploaded Successfully", 
                   "Proceed with QC plot visualization", 
                   type = "success", confirmButtonCol = "#1c82e0")
        
      } else {
        
        withProgress(message = paste("Uploading Data..."), value = 0, {
          
          # Load CCN files
          incProgress(1/4, detail = "Reading example file...")
          ccn_data_files <- readRDS('data/exampleData_scDNAseq.rds')
          
          incProgress(2/4, detail = "Processing example data...")
          ginkgo_data_m <-  as.matrix(ccn_data_files@assays[["ginkgo_data"]]$counts) %>%
            as.data.frame() %>%
            pivot_longer(cols = everything(), names_to = "variable", values_to = "value")
          cl_names <- gsub("_trimmed$", "", ginkgo_data_m$variable)
          ginkgo_data_m$sample <- gsub("_[^_]*$", "", cl_names)        
          ginkgo_data_normalized <- sweep(as.matrix(ccn_data_files@assays[["ginkgo_data"]]$counts) + 1, 2, colMeans(ccn_data_files@assays[["ginkgo_data"]]$counts + 1), '/')
          ginkgo_data_normalized <- cbind(ccn_data_files[["RNA"]][["GC_Content"]], ginkgo_data_normalized)
          ginkgo_data_normalized_m <- as.matrix(ginkgo_data_normalized) %>%
            as.data.frame() %>%
            pivot_longer(cols = -GC_Content, names_to = "variable", values_to = "value")
          cl_names <- gsub("_trimmed$", "", ginkgo_data_normalized_m$variable)
          ginkgo_data_normalized_m$sample <- gsub("_[^_]*$", "", cl_names)
          
          incProgress(3/4, detail = "Calculating Gini scores...")
          gini_scores <- apply(ccn_data_files@assays[["ginkgo_data"]]$counts, MARGIN=2, FUN=Gini)
          cl_names <- gsub("_trimmed$", "", names(gini_scores))
          sample <- gsub("_[^_]*$", "", cl_names)
          gini_scores_df <- data.frame(sample=sample, gini=gini_scores)
          
          # Update data
          incProgress(4/4, detail = "Finalizing data upload...")
          data(list(ccn_data_files = ccn_data_files, 
                    gini_scores = gini_scores_df, 
                    normalized_data_melted = ginkgo_data_normalized_m, 
                    ginkgo_data_melted = ginkgo_data_m, 
                    normalized_data = ginkgo_data_normalized))
          
        })
        
        shinyalert("Example Data Loaded Successfully", 
                   "Proceed with QC plot visualization", 
                   type = "success", confirmButtonCol = "#1c82e0")
      }
      
    })
    return(data)
  })
}