#' Server component for plotting various QC plots
#' 
#' Implements server side logic for showing various QC plots like loess, lorenz, MAD plot etc. 
#' 
#' @param id Unique identifier for the server module, ensuring isolated scope
#'           within the Shiny application.
#' @param ccnObject Object consisting of GC content, raw data, SegCopy, bounds data (obtained from Ginkgo)
#'
#' @return
#' @export
#'
#' @examples
#' GinkgoServer("ginkgoModuleUI", data)
GinkgoServer <- function(id, ccnObject) {
  moduleServer(id, function(input, output, session) {
    
    ns <- session$ns
    data <- reactiveVal(NULL)
    
    shinyjs::hide("download_plot")
    shinyjs::hide("plot_file_type")
    
    # Filtering
    gini_scores_df_filt <- reactive({
      
      req(ccnObject()$gini_scores, input$niqrs)
      n_iqrs <- input$niqrs
      
      is_outlier <- function(x, top_only=TRUE) {
        if (top_only) {
          return(x > quantile(x, 0.75, na.rm = TRUE) + n_iqrs * IQR(x, na.rm = TRUE))
        } else {
          return(x < quantile(x, 0.25, na.rm = TRUE) - n_iqrs * IQR(x, na.rm = TRUE) | x > quantile(x, 0.75, na.rm = TRUE) + n_iqrs * IQR(x, na.rm = TRUE))
        }
      }
      
      cell_ids <- rownames(ccnObject()$gini_scores)
      gini_scores_df <- as.data.frame(ccnObject()$gini_scores %>% group_by(sample) %>% mutate(outlier = ifelse(is_outlier(gini), TRUE, FALSE)))
      rownames(gini_scores_df) <- cell_ids
      cells_to_remove <- rownames(subset(gini_scores_df, outlier == TRUE))
      outlier_table <- table(subset(gini_scores_df, outlier == TRUE)$sample)
      list(gini_scores_with_outliers = gini_scores_df, cells_to_remove = cells_to_remove)
    })
    
    # CCN data
    ccn_data <- reactive({
      req(ccnObject(), gini_scores_df_filt()$cells_to_remove)
      ccn_medians <- data.frame(ploidy=colMedians(as.matrix(ccnObject()$ccn_data_files@assays[["RNA"]]$counts)))
      cl_names <- gsub("_trimmed$", "", rownames(ccn_medians))
      ccn_medians$sample <- as.factor(gsub("_[ACTG]+$", "", cl_names))
      ccn_medians$barcode <- rownames(ccn_medians)
      ccn_medians
    })
    
    # Merge ccn data with gini scores dataframe
    merged_data <- reactive({
      req(gini_scores_df_filt()$gini_scores_with_outliers, ccn_data())
      df <- gini_scores_df_filt()$gini_scores_with_outliers
      df$barcode <- rownames(df)
      # Merging
      merged <- merge(df, ccn_data(), by = "barcode")
      merged <- merged[,!names(merged) %in%  "sample.y"]
      colnames(merged)[which(names(merged) == "sample.x")] <- "sample"
      merged
    })
    
    # Dynamically populate choices for sample, ploidy, and barcode based on merged data
    observe({
      updateSelectInput(session, "sample", choices = unique(merged_data()$sample))
      
      # Update ploidy slider range based on data
      updateSliderInput(session, "ploidy", min = min(merged_data()$ploidy, na.rm = TRUE),
                        max = max(merged_data()$ploidy, na.rm = TRUE),
                        value = c(min(merged_data()$ploidy, na.rm = TRUE), 
                                  max(merged_data()$ploidy, na.rm = TRUE)))
    })
    
    # Update barcodes dynamically based on the selected sample and ploidy filters
    observe({
      filtered_data <- merged_data()
      
      # Apply sample filter
      if (!is.null(input$sample) && length(input$sample) > 0) {
        filtered_data <- filtered_data %>% filter(sample %in% input$sample)
      }
      
      # Apply ploidy filter based on slider range
      filtered_data <- filtered_data %>% filter(ploidy >= input$ploidy[1], 
                                                ploidy <= input$ploidy[2])
      
      # Update barcode choices
      updateSelectizeInput(session, "barcode", choices = unique(filtered_data$barcode), server = TRUE)
    })
    
    # Reactive expression to filter the merged data based on user input for sample, ploidy, and barcode
    filtered_merged_data <- reactive({
      data <- merged_data()
      # Apply sample filter
      if (!is.null(input$sample) && length(input$sample) > 0) {
        data <- data %>% filter(sample %in% input$sample)
      }
      
      # Apply ploidy filter
      data <- data %>% filter(ploidy >= input$ploidy[1], 
                              ploidy <= input$ploidy[2])
      
      # Apply barcode filter
      if (!is.null(input$barcode) && length(input$barcode) > 0) {
        data <- data %>% filter(barcode %in% input$barcode)
      }
      data
    })
    
    # MAD Scores
    mad_df <-  reactive({
      req(ccnObject()$normalized_data, gini_scores_df_filt()$cells_to_remove)
      off_by <- 1
      mad_scores <- apply(ccnObject()$normalized_data[-(1:off_by), 2:ncol(ccnObject()$normalized_data)] - ccnObject()$normalized_data[1:(nrow(ccnObject()$normalized_data)-off_by), 2:ncol(ccnObject()$normalized_data)], MARGIN=2, FUN=mad)
      mad_scores_df <- data.frame(mad=mad_scores)
      mad_scores_df$barcode <- rownames(mad_scores_df)
      
      cl_names <- gsub("_trimmed$", "", rownames(mad_scores_df))
      mad_scores_df$sample <- gsub("_[^_]*$", "", cl_names)
      
      `%notin%` <- Negate(`%in%`)
      mad_scores_df <- subset(mad_scores_df, barcode %notin% gini_scores_df_filt()$cells_to_remove)
      mad_scores_m <- merge(mad_scores_df, ccn_data(), by = "barcode")
      mad_scores_m <- mad_scores_m[,!names(mad_scores_m) %in%  "sample.y"]
      colnames(mad_scores_m)[which(names(mad_scores_m) == "sample.x")] <- "sample"
      mad_scores_m
    })
    
    # Apply filters for other plots
    apply_filters <- function(data, sample_input, ploidy_input, barcode_input) {
      if (!is.null(sample_input) && length(sample_input) > 0) {
        data <- data %>% filter(sample %in% sample_input)
      }
      
      data <- data %>% filter(ploidy >= input$ploidy[1], 
                              ploidy <= input$ploidy[2])
      
      if (!is.null(barcode_input) && length(barcode_input) > 0) {
        data <- data %>% filter(barcode %in% barcode_input)
      }
      
      return(data)
    }
    
    # Loess data reactive expression
    loess_data <- reactive({
      req(ccnObject())
      loess_datam <- ccnObject()$normalized_data_melted
      colnames(loess_datam)[2] <- "barcode"
      loess_datam <- merge(loess_datam, ccn_data(), by = "barcode")
      loess_datam <- loess_datam %>% 
        select(-sample.y) %>% 
        rename(sample = sample.x)
      loess_datam
    })
    
    # Lorenz data reactive expression
    lorenz_data <- reactive({
      req(ccnObject())
      ginkgo_data <- ccnObject()$ginkgo_data_melted
      colnames(ginkgo_data)[1] <- "barcode"
      lorenz_datam <- merge(ginkgo_data, ccn_data(), by = "barcode")
      lorenz_datam <- lorenz_datam %>% 
        select(-sample.y) %>% 
        rename(sample = sample.x)
      lorenz_datam
    })
    
    # Define the plotting functions without reactives
    gini_plot <- reactive({
      withProgress(message = 'Generating Gini Plot', value = 0, {
        incProgress(0.1, detail = "Generating Gini Plot...")
        p <- ggplot(filtered_merged_data(), aes(x = sample, y = gini, color = sample)) +
          geom_boxplot() +
          geom_point(data = subset(filtered_merged_data(), outlier == TRUE), aes(x = sample, y = gini), color = "black", pch = 4) +
          ylim(0, 1) +
          geom_hline(yintercept = 0.5, linetype = "dashed") +
          labs(title = "Coverage Uniformity", x = "Sample", y = "Gini Coefficient") +
          theme_minimal() +
          theme(legend.position = "none")
        incProgress(0.5, detail = "Finalizing Gini Plot...")
        return(p)
      })
    })
    
    loess_plot <- reactive({
      withProgress(message = 'Generating Loess Plot', value = 0, {
        incProgress(0.1, detail = "Processing Loess plot data...")
        loess_data_m <- apply_filters(loess_data(), input$sample, input$ploidy, input$barcode)
        `%notin%` <- Negate(`%in%`)
        subset_data <- subset(loess_data_m, barcode %notin% gini_scores_df_filt()$cells_to_remove)
        incProgress(0.5, detail = "Rendering Loess plot...")
        p <- ggplot(subset_data, aes(x = GC_Content, y = log(value), group = barcode)) +
          geom_smooth(method = "loess", se = FALSE, span = 0.05, linewidth = 0.1, color = rgb(28, 130, 224, maxColorValue = 255)) +
          xlim(0.3, NA) +
          ylim(-5, NA) +
          labs(title = "GC Content vs Bin Counts", x = "GC Content", y = "Log of Normalized Read Counts") +
          theme_minimal() +
          theme(legend.position = "none")
        incProgress(1, detail = "Finalizing Loess Plot...")
        return(p)
      })
    })
    
    lorenz_plot <- reactive({
      withProgress(message = 'Generating Lorenz Plot', value = 0, {
        incProgress(0.1, detail = "Processing Lorenz plot data...")
        lorenz_data_m <- apply_filters(lorenz_data(), input$sample, input$ploidy, input$barcode)
        `%notin%` <- Negate(`%in%`)
        incProgress(0.5, detail = "Rendering Lorenz plot...")
        subset_data_for_lorenz <- subset(lorenz_data_m, barcode %notin% gini_scores_df_filt()$cells_to_remove)
        p <- ggplot(subset_data_for_lorenz, aes(x = value, group = barcode)) +
          stat_lorenz(color = rgb(28, 130, 224, maxColorValue = 255), linewidth = 0.1) +
          coord_fixed() +
          geom_abline(linetype = "dashed") +
          labs(title = "Lorenz Curve of Coverage Uniformity", x = "Cumulative Fraction of Genome", y = "Cumulative Fraction of Total Reads") +
          theme_minimal()
        incProgress(1, detail = "Finalizing Lorenz Plot...")
        return(p)
      })
    })
    
    mad_plot_by_samples <- reactive({
      withProgress(message = 'Generating MAD Plot by Samples', value = 0, {
        incProgress(0.1, detail = "Processing MAD plot data by samples...")
        mad_df_m <- apply_filters(mad_df(), input$sample, input$ploidy, input$barcode)
        p <- ggplot(mad_df_m, aes(x = sample, color = sample, y = mad)) +
          geom_boxplot() +
          ylim(0, 1) +
          labs(title = "Mean Absolute Deviation\nof Neighboring Bins (By Samples)", x = "Sample", y = "Mean Absolute Deviation (MAD)") +
          theme_minimal() +
          theme(legend.position = "none")
        incProgress(1, detail = "Finalizing MAD Plot by Samples...")
        return(p)
      })
    })
    
    mad_plot <- reactive({
      withProgress(message = 'Generating MAD Plot', value = 0, {
        incProgress(0.1, detail = "Processing MAD plot data...")
        mad_df_m <- apply_filters(mad_df(), input$sample, input$ploidy, input$barcode)
        p <- ggplot(mad_df_m, aes(y = mad)) +
          geom_boxplot(color = rgb(28, 130, 224, maxColorValue = 255)) +
          ylim(0, 1) +
          labs(title = "Mean Absolute Deviation\nof Neighboring Bins", y = "Mean Absolute Deviation (MAD)") +
          theme_minimal() +
          theme(legend.position = "none", axis.title.x = element_blank(), axis.text.x = element_blank(),
                axis.ticks.x = element_blank())
        incProgress(1, detail = "Finalizing MAD Plot...")
        return(p)
      })
    })
    
    plot_data <- reactiveVal()
    
    # Generate plot and save it to a temporary file
    observeEvent(input$plotButton, {
      
      if (is.null(ccnObject())) {
        shinyalert('Missing Data', 'Please upload data.', type = "error", confirmButtonCol = "#1c82e0")
        return(NULL)
      } 
      
      gini_scores_filtered <- gini_scores_df_filt()
      
      # Alert if no data present based on user selection
      if (all(gini_scores_filtered$gini_scores_with_outliers$outlier==FALSE)) {
        shinyalert("No outliers detected",
                   "The outlier detection threshold did not detect any outliers. Consider adjusting the threshold.",
                   type = "error", confirmButtonCol = "#1c82e0")
        return(NULL)
      }
      
      # Parameter tracking
      updated_ccnObject <- ccnObject()
      updated_ccnObject$user_parameter_tracker <- NULL
      updated_ccnObject$user_parameter_tracker <- c(
        updated_ccnObject$user_parameter_tracker,
        paste("Outlier detection threshold (nIQRs):", input$niqrs),
        paste("Selected Sample(s):", paste(input$sample, collapse = ", ")),
        paste("Selected Ploidy:", paste(input$ploidy, collapse = ", ")),
        paste("Selected Barcode(s):", paste(input$barcode, collapse = ", "))
      )
      
      # Update the object
      data(updated_ccnObject) # update data
      
      plot_to_save <- switch(input$plot_type,
                             "Gini_Plot" = gini_plot(),
                             "Loess_Plot" = loess_plot(),
                             "lorenz_plot" = lorenz_plot(),
                             "MAD_Plot_by_samples" = mad_plot_by_samples(),
                             "MAD_plot" = mad_plot())
      
      plot_data(plot_to_save)  # Update the plot
    })
    
    
    # Render plot
    output$plot <- renderPlot({
      req(plot_data())
      plot_data()
    })
    
    # Show the download button based on the plot existence
    observe({
      if (!is.null(plot_data())) {
        shinyjs::show("download_plot")
        shinyjs::show("plot_file_type")
      }
    })
    
    # Download handler to save the plot
    output$download_plot <- downloadHandler(
      filename = function() {
        paste0("CogentDS_", input$plot_type, "_", Sys.Date(), input$plot_file_type)
      },
      content = function(file) {
        req(plot_data())
        withProgress(message = "Saving plot...", value = 0, {
          incProgress(0.5, detail = "Generating plot file...")
          # Save the plot with the progress indicator
          ggsave(file, plot = plot_data(), width = 4, height = 4, units = "in", bg = "white")
          incProgress(0.2, detail = "Finalizing save...")
        })
      },
      contentType = 'image'
    )
    return(data)
  })
}