#' CN Profile Plot Server Component
#' 
#' Implement server side logic for showing CN profile for selected barcodes
#'
#' @param id Unique identifier for the server module, ensuring isolated scope
#'           within the Shiny application.
#'           
#' @param ccnObject Object consisting of GC content, raw data, SegCopy, bounds data (obtained from Ginkgo)
#'
#' @return
#' @export
#'
#' @examples
#' ccnplotServer("ccnplotModuleUI", data)
ccnplotServer <- function(id, ccnObject) {
  moduleServer(id, function(input, output, session) {
    
    ns <- session$ns
    
    shinyjs::hide("download_plot")
    shinyjs::hide("plot_file_type")
    
    ccn_plot <- reactiveVal(NULL)
    
    # Update barcodes 
    observe({
      req(ccnObject())
      barcodes <- colnames(as.matrix(SeuratObject::GetAssayData(ccnObject()$ccn_data_files[["RNA"]], layer = "counts")))
      # Update barcode choices, allowing up to 3 selections
      updateSelectizeInput(session, "barcode", choices = unique(barcodes), server = TRUE, options = list(maxItems = 3))
    })
    
    # Bounds and pos reactives
    bounds <- reactive({
      req(ccnObject())
      ccnObject()$ccn_data_files@misc$bounds
    })
    
    pos <- reactive({
      req(bounds())
      seg_copy <- as.matrix(GetAssayData(ccnObject()$ccn_data_files[["RNA"]], layer = "counts"))
      l <- nrow(seg_copy)
      cbind(c(1, bounds()[,2]), c(bounds()[,2], l))
    })
    
    # Data processing based on selected samples
    precomputedData <- reactive({
      
      selected_samples <- input$barcode  # Handle up to 3 selected samples
      if (length(selected_samples) == 0) {
        return(NULL)
      }
      
      req(ccnObject())
      
      # Get necessary data files
      raw <- as.matrix(GetAssayData(ccnObject()$ccn_data_files[["ginkgo_data"]], layer = "counts"))
      GC <- data.frame(GC = ccnObject()$ccn_data_files[["RNA"]][["GC_Content"]])
      ccn_chr_info <- ccnObject()$ccn_data_files[["RNA"]][["Chrom_info"]] 
      loc <- separate(ccn_chr_info, Chrom_info, into = c("CHR", "START", "END"), sep = "-")
      
      # Variables and initialization
      minPloidy <- 1.5
      maxPloidy <- 6
      minBinWidth <- 5
      
      ploidy <- rbind(c(0, 0), c(0, 0))
      l <- nrow(raw)
      w <- length(selected_samples)  # Only process selected samples 
      CNgrid <- seq(minPloidy, maxPloidy, by = 0.05)
      n_ploidy <- length(CNgrid)
      CNmult <- matrix(0, n_ploidy, w)
      CNerror <- matrix(0, n_ploidy, w)
      outerColsums <- matrix(0, n_ploidy, w)
      fixed <- matrix(0, l, w)
      breaks <- matrix(0, l, w)
      normal <- sweep(raw + 1, 2, colMeans(raw + 1), '/')
      
      lab <- colnames(normal)
      best_pred_CN <- numeric(w)
      
      # Define the GC correction function
      lowess.gc <- function(jtkx, jtky) {
        jtklow <- lowess(jtkx, log(jtky), f = 0.05)
        jtkz <- approx(jtklow$x, jtklow$y, jtkx)
        return(exp(log(jtky) - jtkz$y))
      }
      
      withProgress(message = "Plotting CN profile", value = 0, {
        
        # Process only the selected samples
        for (k in seq_along(selected_samples)) {
          selected_sample_index <- which(lab == selected_samples[k])
          
          # Correct for GC bias
          normal[, selected_sample_index] <- lowess.gc(GC[, 1], (raw[, selected_sample_index] + 1) / mean(raw[, selected_sample_index] + 1))
          
          # Calculate log ratio
          lr <- log2(normal[, selected_sample_index])
          
          # Segmentation using DNAcopy
          CNA.object <- CNA(genomdat = lr, chrom = loc[, 1], maploc = as.numeric(loc[, 2]), data.type = 'logratio')
          CNA.smoothed <- smooth.CNA(CNA.object)
          segs <- segment(CNA.smoothed, verbose = 0, min.width = minBinWidth)
          frag <- segs$output[, 2:3]
          
          # Identify breakpoints
          len <- nrow(frag)
          bps <- numeric(len)
          for (j in 1:len) {
            bps[j] <- which((loc[, 1] == frag[j, 1]) & (as.numeric(loc[, 2]) == frag[j, 2]))
          }
          bps <- sort(bps)
          bps[len + 1] <- l
          breaks[bps, k] <- 1
          
          # Adjust segments
          fixed[1:bps[2], k] <- median(normal[1:bps[2], selected_sample_index])
          for (i in 2:len) {
            fixed[bps[i]:(bps[i + 1] - 1), k] <- median(normal[bps[i]:(bps[i + 1] - 1), selected_sample_index])
          }
          fixed[, k] <- fixed[, k] / mean(fixed[, k])
          
          # Determine Copy Number using SoS method
          outerRaw <- fixed[, k] %o% CNgrid
          outerRound <- round(outerRaw)
          outerDiff <- (outerRaw - outerRound) ^ 2
          outerColsums[, k] <- colSums(outerDiff)
          CNmult[, k] <- CNgrid[order(outerColsums[, k])]
          CNerror[, k] <- round(sort(outerColsums[, k]), 2)
          
          # Select CN based on conditions
          if (length(which(lab[selected_sample_index] == ploidy[, 1])) == 0) {
            CN <- CNmult[1, k]
          } else {
            estimate <- ploidy[which(lab[selected_sample_index] == ploidy[, 1]), 2]
            CN <- CNmult[which(abs(CNmult[, k] - estimate) < 0.4), k][1]
          }
          best_pred_CN[k] <- CN
          
          incProgress(1 / w, detail = paste("Processing samples..."))
        }
      })
      
      list(normal = normal, best_pred_CN = best_pred_CN)
    })
    
    observeEvent(input$plot, {
      
      if (is.null(ccnObject())) {
        shinyalert('Missing Data', 'Please upload the data.', type = "error", confirmButtonCol = "#1c82e0")
        return(NULL)
      } 
      
      req(ccnObject())
      
      precomputed <- precomputedData()
      
      final <- as.matrix(ccnObject()$ccn_data_files@assays[["RNA"]]$counts)
      best_pred_CN <- precomputed$best_pred_CN
      pos_data <- pos()
      bounds_data <- bounds()
      lab <- colnames(final)  
      selected_samples <- input$barcode
      
      if (length(selected_samples) == 0) {
        shinyalert("Please select at least 1 sample.", type = "error", confirmButtonCol = "#1c82e0")
        return(NULL)
      }
      
      top <- 8
      
      plot_list <- lapply(1:length(selected_samples), function(i) {
        selected_sample <- which(lab == selected_samples[i])
        CN <- best_pred_CN[i]
        
        rectangles1 <- data.frame(xmin = pos_data[seq(1, nrow(pos_data), 2), 1], xmax = pos_data[seq(1, nrow(pos_data), 2), 2])
        rectangles2 <- data.frame(xmin = pos_data[seq(2, nrow(pos_data), 2), 1], xmax = pos_data[seq(2, nrow(pos_data), 2), 2])
        clouds <- data.frame(x = 1:nrow(final), y = precomputed$normal[, selected_sample] * CN)
        amp <- data.frame(x = which(final[, selected_sample] > 2), y = final[which(final[, selected_sample] > 2), selected_sample])
        del <- data.frame(x = which(final[, selected_sample] < 2), y = final[which(final[, selected_sample] < 2), selected_sample])
        flat <- data.frame(x = which(final[, selected_sample] == 2), y = final[which(final[, selected_sample] == 2), selected_sample])
        anno <- data.frame(x = (pos_data[, 2] + pos_data[, 1]) / 2, y = -top * .05, chrom = substring(c(as.character(bounds_data[, 1]), "chrY"), 4, 5))
        
        ggplot() +
          geom_rect(data = rectangles1, aes(xmin = xmin, xmax = xmax, ymin = -top * .1, ymax = top), fill = 'gray85', alpha = 0.75) +
          geom_rect(data = rectangles2, aes(xmin = xmin, xmax = xmax, ymin = -top * .1, ymax = top), fill = 'gray75', alpha = 0.75) +
          geom_point(data = clouds, aes(x = x, y = y), color = 'gray45', size = 0.5) +
          geom_point(data = flat, aes(x = x, y = y), size = 0.5) +
          geom_point(data = amp, aes(x = x, y = y), size = 0.5, color = 'red') +
          geom_point(data = del, aes(x = x, y = y), color = 'blue') +
          geom_text(data = anno, aes(x = x, y = y, label = chrom), size = 6) +
          scale_x_continuous(limits = c(0, nrow(final)), expand = c(0, 0)) +
          scale_y_continuous(limits = c(-top * .1, top), expand = c(0, 0)) +
          labs(title = paste("Integer Copy Number Profile for Sample \"", lab[selected_sample], "\"\n Predicted Ploidy = ", CN, sep = ""), 
               x = "Chromosome", y = "Copy Number", size = 14) +
          theme(plot.title = element_text(size = 16, vjust = 1.5)) +
          theme(axis.title.x = element_text(size = 16, vjust = -0.05), axis.title.y = element_text(size = 16, vjust = 0.1)) +
          theme(axis.text = element_text(color = "black", size = 16), axis.ticks = element_line(color = "black")) +
          theme(axis.ticks.x = element_blank(), axis.text.x = element_blank(), axis.line.x = element_blank()) +
          theme(panel.background = element_rect(fill = 'gray90')) +
          theme(plot.margin = unit(c(0.5, 1, 0.5, 1), "cm")) +
          theme(panel.grid.major.x = element_blank()) +
          geom_vline(xintercept = c(1, nrow(final)), size = 0.5) +
          geom_hline(yintercept = c(-top * 0.1, top), size = 0.5)
      })
      
      combined_plot <- patchwork::wrap_plots(plot_list, ncol = 1)
      ccn_plot(combined_plot)
      
      # Dynamic UI for plot based on number of selected samples
      output$plotUI <- renderUI({
        plot_height <- if (length(selected_samples) == 1) "400px" else "700px"
        shinycssloaders::withSpinner(plotOutput(ns("CCNPlot"), width = "100%", height = plot_height))
      })
    })
    
    output$CCNPlot <- renderPlot({
      req(ccn_plot())
      ccn_plot()
    })
    
    observe({
      if (!is.null(ccn_plot())) {
        shinyjs::show("download_plot")
        shinyjs::show("plot_file_type")
      }
    })
    
    # Download handler with consistent dimensions
    output$download_plot <- downloadHandler(
      filename = function() {
        paste0("CogentDS_", "CCN_Profile", "_", Sys.Date(), input$plot_file_type)
      },
      content = function(file) {
        req(ccn_plot())
        withProgress(message = "Preparing plot for download...", value = 0, {
        
        # Adjust width and height based on the number of selected samples
        plot_width <- 25  
        plot_height <- if (length(input$barcode) == 1) 6 else 12
        Sys.sleep(0.5)
        ggsave(file, plot = ccn_plot(), width = plot_width, height = plot_height)
        # Finalize progress
        incProgress(0.2, detail = "Finalizing download...")
      })
      },
      contentType = 'image'
    )
    
  })
}