#' Report generation Server component
#' 
#' Implement server side logic for analysis report generation
#'
#' @param id Unique identifier for the server module, ensuring isolated scope
#'           within the Shiny application.
#' @param ccnObject Object consisting of parameters from QC module
#' @param heatmap_data Object consisting of parameters from Heatmap module
#' @param umap_data Object consisting of parameters from Clustering module
#'
#' @return
#' @export
#'
#' @examples
#' generateReport_Server("generateReport_Module", qcobject, heatmap_data, umap_data$user_param_tracker_clustering)
generateReport_Server <- function(id, ccnObject, heatmap_data, umap_data) {
  
  moduleServer(id, function(input, output, session) {
    ns <- session$ns
    
    output$scDNA_Report <- downloadHandler(
      filename = function() {
      paste0("CogentDS_Analysis_Report", "_", Sys.Date(), ".html")
      },
      content = function(file) {
        withProgress(message = 'Preparing Report...', value = 0, {
          
          if (is.null(ccnObject())) {
            shinyalert('Missing Data', 'Please run the QC module.', type = "error", confirmButtonCol = "#1c82e0")
            return(NULL)
          } 
          if (is.null(heatmap_data())) {
            shinyalert('Missing Data', 'Please run the Heatmap module.', type = "error", confirmButtonCol = "#1c82e0")
            return(NULL)
          }
          if (is.null(umap_data())) {
            shinyalert('Missing Data', 'Please run the UMAP module.', type = "error", confirmButtonCol = "#1c82e0")
            return(NULL)
          }
          
          incProgress(0.3, detail = "Generating Report elements ...")
          tryCatch({
            cnv_data <- ccnObject()
            heatmap_params <- heatmap_data()
            clustering_params <- umap_data()
            Report_rmd <- paste0('R/Setup_Funtions/bin/', 'CogentDS_scDNA_analyis_report_App.Rmd')
            params <- list(set_title = "Cogent™ NGS Discovery Software", 
                           seuratObject_gene_preprocessed = cnv_data, 
                           heatmap_params =heatmap_params,
                           clustering_params = clustering_params)
            
            message("Rendering report...")
            rmarkdown::render(Report_rmd,
                              output_file = file,
                              params = params,
                              envir = new.env(parent = globalenv()))
            shinyjs::disable("scDNA_Report")
            shinyalert('Analysis Report Downloaded', "Thank you! CogentDS Team", type = "success", confirmButtonCol = "#1c82e0")
          }, error = function(e) {
            
            shinyalert('Error in downloading the analysis report', 
                       'Please check your input parameters.', type = "error", confirmButtonCol = "#1c82e0")
          })
          incProgress(0.7, detail = "Completed")
        })
      } 
    )
  })
}