#'  Server logic for Lasso selection
#'
#' Implements server side functionality for UMAP lasso selection. It includes handling
#' of lasso selection on the UMAP plot based on samples/clusters
#' 
#' @param id Unique identifier for the server module, ensuring isolated scope
#'           within the Shiny application.
#' @param umap_data UMAP cooridnates obtained from running clustering module
#' @param ccnObject Object consisting of GC content, raw data, SegCopy, bounds data (obtained from Ginkgo)
#'
#' @return
#' @export
#'
#' @examples
#' lassoServer("lassoUIModule", umap_data, data)
lassoServer <- function(id, umap_data, ccnObject) {
  moduleServer(id, function(input, output, session) {
    
    ns <- session$ns
    
    # Reactive for holding selections
    combined_selection <- reactiveVal()
    
    # Data preparation for UMAP
    plot_data <- reactiveVal(NULL)
    
    # Filtering based on gini scores
    gini_scores_df_filt <- reactive({
      req(ccnObject()$gini_scores, input$niqrs)
      n_iqrs <- input$niqrs
      
      is_outlier <- function(x, top_only=TRUE) {
        if (top_only) {
          return(x > quantile(x, 0.75, na.rm = TRUE) + n_iqrs * IQR(x, na.rm = TRUE))
        } else {
          return(x < quantile(x, 0.25, na.rm = TRUE) - n_iqrs * IQR(x, na.rm = TRUE) | x > quantile(x, 0.75, na.rm = TRUE) + n_iqrs * IQR(x, na.rm = TRUE))
        }
      }
      
      cell_ids <- rownames(ccnObject()$gini_scores)
      gini_scores_df <- as.data.frame(ccnObject()$gini_scores %>% group_by(sample) %>% mutate(outlier = ifelse(is_outlier(gini), TRUE, FALSE)))
      rownames(gini_scores_df) <- cell_ids
      cells_to_remove <- rownames(subset(gini_scores_df, outlier == TRUE))
      list(gini_scores_with_outliers = gini_scores_df, cells_to_remove = cells_to_remove)
    })
    
    # CCN data
    ccn_data <- reactive({
      req(ccnObject(), gini_scores_df_filt()$cells_to_remove)
      ccn_medians <- data.frame(ploidy=colMedians(as.matrix(ccnObject()$ccn_data_files@assays[["RNA"]]$counts)))
      cl_names <- gsub("_trimmed$", "", rownames(ccn_medians))
      ccn_medians$sample <- as.factor(gsub("_[ACTG]+$", "", cl_names))
      ccn_medians$barcode <- rownames(ccn_medians)
      ccn_medians
    })
    
    # Merge ccn data with gini scores dataframe
    merged_data <- reactive({
      req(gini_scores_df_filt()$gini_scores_with_outliers, ccn_data())
      df <- gini_scores_df_filt()$gini_scores_with_outliers
      df$barcode <- rownames(df)
      # Merging
      merged <- merge(df, ccn_data(), by = "barcode")
      merged <- merged[,!names(merged) %in%  "sample.y"]
      colnames(merged)[which(names(merged) == "sample.x")] <- "sample"
      merged
    })
    
    # Dynamically populate choices for sample, ploidy, and barcode based on merged data
    observe({
      updateSelectInput(session, "sample", choices = unique(plot_data()$sample))
      # Update ploidy slider range based on data
      updateSliderInput(session, "ploidy", min = min(merged_data()$ploidy, na.rm = TRUE),
                        max = max(merged_data()$ploidy, na.rm = TRUE),
                        value = c(min(merged_data()$ploidy, na.rm = TRUE), 
                                  max(merged_data()$ploidy, na.rm = TRUE)))    })
    
    # Update barcodes dynamically based on the selected sample and ploidy filters
    observe({
      req(plot_data())
      filtered_data <- plot_data()
      
      # Apply sample filter
      if (!is.null(input$sample) && length(input$sample) > 0) {
        filtered_data <- filtered_data %>% filter(sample %in% input$sample)
      }
      
      # Apply ploidy filter
      filtered_data <- filtered_data %>% filter(ploidy >= input$ploidy[1], 
                                                ploidy <= input$ploidy[2])
      
      # Update barcode choices
      updateSelectizeInput(session, "barcode", choices = unique(filtered_data$barcode), server = TRUE)
    })
    
    # Data preparation for lasso
    observe({
      req(umap_data(), merged_data())
      udata <- umap_data()
      
      # Initialize custom_selection if not already present
      if (!"custom_clusters" %in% colnames(udata)) {
        udata$custom_clusters <- as.character(udata$cluster)
      }
      
      # Initialize custom_selection if not already present
      if (!"custom_samples" %in% colnames(udata)) {
        udata$custom_samples <- as.character(udata$sample)
      }
      
      # Merge umap coordinates with metadata info to include ploidy info
      udata <- merge(udata, merged_data(), by = "barcode")
      udata <- udata[,!names(udata) %in%  "sample.y"]
      colnames(udata)[which(names(udata) == "sample.x")] <- "sample"
      
      # Store initial UMAP data in the reactive value
      plot_data(udata)
    })
    
    # Filter plot data based on user selections for sample, ploidy, and barcode
    filtered_plot_data <- reactive({
      req(plot_data())
      data <- plot_data()
      
      # Apply sample filter
      if (!is.null(input$sample) && length(input$sample) > 0) {
        data <- data %>% filter(sample %in% input$sample)
      }
      
      # Apply ploidy filter
      data <- data %>% filter(ploidy >= input$ploidy[1], 
                              ploidy <= input$ploidy[2])
      
      # Apply barcode filter
      if (!is.null(input$barcode) && length(input$barcode) > 0) {
        data <- data %>% filter(barcode %in% input$barcode)
      }
      
      if (!is.null(gini_scores_df_filt()$cells_to_remove)) {
        data <- data %>% filter(!barcode %in% gini_scores_df_filt()$cells_to_remove)
      }
      
      data
    })
    
    # Render plots based on the user select plot type from dropdown
    output$UMAP_plot <-  renderPlotly({
      
      # Ensure gini_scores_df_filt gets re-evaluated based on the current outlier threshold value
      gini_scores_filtered <- gini_scores_df_filt()
      
      # Alert if no data present based on user selection
      if (all(gini_scores_filtered$gini_scores_with_outliers$outlier==FALSE)) {
        shinyalert("No outliers detected",
                   "The outlier detection threshold did not detect any outliers. Consider adjusting the threshold.",
                   type = "error", confirmButtonCol = "#1c82e0")
        return(NULL)
      }
      
      if (is.null(ccnObject())) {
        shinyalert('Missing Data', 'Please upload the data.', type = "error", confirmButtonCol = "#1c82e0")
        return(NULL)
      } 
      
      if (input$color_by == "sample" && !"sample" %in% colnames(filtered_plot_data())) {
        shinyalert("Error", "Sample not available. Please run clustering module first.", type = "error", confirmButtonCol = "#1c82e0")
        return(NULL)
      }
      
      if (input$color_by == "cluster" && !"cluster" %in% colnames(filtered_plot_data())) {
        shinyalert("Error", "Clusters not available. Please run clustering module first.", type = "error", confirmButtonCol = "#1c82e0")
        return(NULL)
      }
      
      req(filtered_plot_data())
      
      color_by <- if (input$color_by == "cluster") {
        "cluster"
      } else {
        "sample"
      }
      
      subset_data <- filtered_plot_data()
      rownames(subset_data) <- subset_data$barcode

      withProgress(message = paste("Generating", "UMAP"), value = 0, {
        pData <- subset_data
        pData$cluster <- as.factor(pData$cluster)
        set.seed(1)
        p <- ggplot(pData, aes(x = UMAP1, y = UMAP2, color = .data[[color_by]])) +
          geom_point(size = 0.5) + theme_classic()
        set.seed(1)
        ggplotly(p, source = "lassoplot") %>% layout(dragmode = 'lasso')
      })
    })
    
    observeEvent(event_data("plotly_selected", source = "lassoplot"), {
      req(filtered_plot_data())
      embedding <- filtered_plot_data()[,c("UMAP1", "UMAP2")]
      embedding$original_index <- 1:nrow(embedding)
      selected_points <- event_data("plotly_selected", source = "lassoplot")
      if (!is.null(selected_points) && length(selected_points) > 0 && !is.null(selected_points$x) && !is.null(selected_points$y)) {
        selected_coords <- selected_points[c("x", "y")]
        tolerance <- 1e-5
        
        matched_indices <- lapply(1:nrow(selected_coords), function(i) {
          which(abs(embedding[, 1] - selected_coords$x[i]) < tolerance & 
                  abs(embedding[, 2] - selected_coords$y[i]) < tolerance)
        })
        
        matched_indices <- unlist(matched_indices)
        if (length(matched_indices) == 0) {
          shinyalert("No points selected", "Please select points on the plot.", type = "warning", confirmButtonCol = "#1c82e0")
        } else {
          valid_indices <- embedding$original_index[matched_indices]
          combined_selection(valid_indices)
          
          # Show modal for naming the selection
          showModal(modalDialog(
            title = "Save Lasso Selection",
            textInput(ns("lassoNameModal"), "Name your selection", ""),
            footer = tagList(
              modalButton("Cancel"),
              actionButton(ns("saveLassoSelectionModal"), "Save", style='color:#fff; background-color:#1c82e0')
            )
          ))
        }
      } else {
        shinyalert("No points selected", "Please select points on the plot.", type = "warning", confirmButtonCol = "#1c82e0")
      }
    })
    
    observeEvent(input$saveLassoSelectionModal, {
      req(combined_selection())
      lassoName <- input$lassoNameModal
      
      if (lassoName == "") {
        shinyalert("Error", "Please provide a name for the lasso selection.", type = "error", confirmButtonCol = "#1c82e0")
      } else {
        
        plot_object <- filtered_plot_data()
        
        print(dim(plot_object))
        
        # Disable inputs after saving lasso selection
        shinyjs::disable("niqrs")
        shinyjs::disable("sample")
        shinyjs::disable("ploidy")
        shinyjs::disable("barcode")
        shinyjs::disable("color_by")
        
        if (input$color_by == "cluster") {
          plot_object$custom_clusters[combined_selection()] <- lassoName
        } else if (input$color_by == "sample") {
          plot_object$custom_samples[combined_selection()] <- lassoName
        }
        
        # Update plot_object
        plot_data(plot_object)
        
        shinyalert("Success", paste("Lasso selection saved as", 
                                    lassoName), 
                   type = "success", confirmButtonCol = "#1c82e0")
        
        # Render plot
        output$UMAP_plot <-  renderPlotly({
          
          color_by <- if (input$color_by == "cluster") {
            "custom_clusters"
          } else {
            "custom_samples"
          }
          
          message("Generating UMAP")
          withProgress(message = paste("Generating", "UMAP"), value = 0, {
            req(plot_object)
            plot_object$custom_clusters <- factor(plot_object$custom_clusters, levels = unique(plot_object$custom_clusters))
            set.seed(1)
            p <- ggplot(plot_object, aes(x=UMAP1, y=UMAP2, color = .data[[color_by]])) +
              geom_point(size = 0.5) + theme_classic()
            set.seed(1)
            ggplotly(p, source = "lassoplot") %>% layout(dragmode = 'lasso')
          })
        })
        # 
        # UI for metadata download
        output$downloadUI <- renderUI({
          if (input$saveLassoSelectionModal > 0){
            downloadButton(ns("downloadMetadata"), "Download metadata", style='color:#fff; background-color:#1c82e0')
          }
        })
        
        output$downloadMetadata <- downloadHandler(
          filename = function() {
            paste("CogentDS_metadata_", Sys.Date(), ".csv", sep="")
          },
          content = function(file) {
            req(plot_data())
            write.csv(plot_data(), file, row.names = F)
          }
        )
        # Remove modal after saving
        removeModal()
      }
    })
  })
}