#' Barcode Rank Plot Server Component
#'  
#' This server module handles the server-side logic for Barcode Rank Plot
#'
#' @param id Unique identifier for the server module, ensuring isolated scope
#'           within the Shiny application.
#'         
#' @export
#'
#' @examples
#' # In the server definition of a Shiny app
#' barcoderankServer("barcoderankModule")


barcoderankServer <- function(id) {
  shiny::moduleServer(id, function(input, output, session) {
    ns <- session$ns
    
    shiny::observeEvent(input$generateKneeplot, {
      shinyjs::disable("generateKneeplot")
      withProgress(message = 'Generating Barcode Rank Plot:', {
        incProgress(0.3, detail = 'reading data')
        req(input$barcode_counts)
        barcode_counts_data = read.csv(input$barcode_counts$datapath, header = FALSE, check.names = FALSE, row.names = 1)
        colnames(barcode_counts_data) = c("Sample", "Total_Reads")
        barcode_counts = barcode_counts_data
        barcode_counts = barcode_counts %>% dplyr::filter(!rownames(barcode_counts) %in% c('Short', 'Unselected', 'Undetermined'))
        barcode_counts = barcode_counts %>% dplyr::select('Total_Reads')
        barcode_counts = t(barcode_counts)
        barcode_counts = as.matrix(barcode_counts)
        incProgress(0.3, detail = 'calculating barcode ranks')
        barcode_counts_ranks = barcodeRanks(barcode_counts)
        barcode_counts_ranks_df = data.frame(barcode_counts_ranks)
        inflection = metadata(barcode_counts_ranks)$inflection
        barcode_counts_ranks_df_filter = barcode_counts_ranks_df %>% dplyr::filter(total > inflection)
        rank_cutoff = round(max(barcode_counts_ranks_df_filter$rank))
        annot <- tibble(inflection = inflection, rank_cutoff = rank_cutoff)
        barcode_counts_ranks_df = arrange(barcode_counts_ranks_df, rank)
        if(nrow(barcode_counts_ranks) > 0) {
          tryCatch({
            plot = ggplot(barcode_counts_ranks_df, aes(rank, total)) + geom_path() + geom_hline(yintercept = inflection, linetype = 2, color = "blue") + geom_vline(xintercept = rank_cutoff, linetype = 2, color = "blue") + scale_x_log10() + scale_y_log10() + labs(y = "Total Reads (log10)", x = "Barcode Ranks (log10)") + geom_text(aes(rank_cutoff, inflection, label = paste(rank_cutoff, "cells", "; Total Reads", inflection)), data = annot, vjust = 1, colour = "red", nudge_x  = -0.28, nudge_y = -0.05)
            incProgress(0.4, detail = 'generating barcode rank plot')
            output$kneeplot_static = renderPlotly({
              req(plot)
              ggplotly(plot)
            })

            output$downloadPlot_knee <- downloadHandler(
              filename = function() {
                paste0("CogentDS_Barcode_Rank_Plot", "-", Sys.Date(), input$kneeplot)
              },
              content = function(file) {
                req(plot)
                plot
                ggsave(filename = file, plot = plot, width = 10, height = 10)
              },
              contentType = "image"
            )

            output$downloadRanks <- downloadHandler(
              filename = function() {
              paste0("CogentDS_Barcode_Ranks", "-", Sys.Date(), ".csv")
              },
              content = function(file) {
                req(barcode_counts_ranks_df)
                barcode_counts_data = barcode_counts_data[rownames(barcode_counts_ranks_df),]
                barcode_counts_ranks_df$Sample = barcode_counts_data$Sample
                write.csv(barcode_counts_ranks_df, file, sep=",", quote=F)
              }
            )
              shinyjs::enable("generateKneeplot")
          }, error = function(e) {
              shinyjs::enable("generateKneeplot")
              shinyalert("Barcode Rank Plot Failed", "Please check your input data", type = "error", confirmButtonCol = "#1c82e0")
          })
        } else {
          shinyalert("Error calculating barcode ranks", "Please check your input data", type = "error", confirmButtonCol = "#1c82e0")
        }
      })
    })
  })
}
