#' Clonotype analysis module server logic
#' 
#' This server module handles overlaying clonotypes on UMAP
#'
#' @param id Unique identifier for the server module, ensuring isolated scope
#'           within the Shiny application.
#'           
#' @param seuratObj A reactive expression containing the Seurat object.
#'
#' @return
#' @export
#'
#' @examples
clonotypeAnalysisServer <- function(id, seuratObj) {
  moduleServer(id, function(input, output, session) {
    
    ns <- session$ns
    shinyjs::hide("ClonotypePlt")
    
    output$downloadPlotUI <- renderUI({
      if (!is.null(static_overlay_plot())) {
        downloadButton(ns('downloadPlot'), 'Save plot', style='color:#fff; background-color:#1c82e0')
      }
    })
    
    # Radio btns for selecting clonotype group
    output$radioButtonsSelection <- renderUI({
      radioButtons(ns("overlay_type"), "Data overlay type:",
                   choices = c("TCRb" = "TCRb", "TCRa" = "TCRa", "TCRdg" = "TCRdg", "BCRh" = "BCRh", "BCRl" = "BCRl"),
                   selected = "TCRb")
    })
    
    # Reactive value to hold metadata columns
    metadataColumns <- reactive({
      req(seuratObj())
      
      # Get the names of metadata columns
      all_metadata <- colnames(seuratObj()@meta.data)
      # Columns that are either factors or characters
      relevant_columns <- all_metadata[sapply(seuratObj()@meta.data[, all_metadata], function(col) is.factor(col) || is.character(col))]
      
      # Columns to be potentially removed
      column_not_needed <- c("Barcode")
      
      # Check and remove columns if they exist
      if (any(column_not_needed %in% relevant_columns)) {
        relevant_columns <- relevant_columns[!(relevant_columns %in% column_not_needed)]
      }
  
      # Return the filtered relevant columns
      relevant_columns
    })
    
    # Determine available clonotypes based on selected assay
    clonotypes <- reactive({
      req(seuratObj(), input$overlay_type)
      
      # Check if the assay is present and contains data'
      if (input$overlay_type %in% names(seuratObj()@assays)){
        selected_assay <- seuratObj()[[input$overlay_type]]$counts
        if (!is.null(selected_assay) && nrow(selected_assay) > 0) {
          clonotype_sums = NULL
          clonotype_sums = rowSums(selected_assay)
          clonotype_sums = data.frame(clonotype_sums)
          colnames(clonotype_sums) = "Value"
          clonotype_sums_present = clonotype_sums %>% dplyr::filter(Value > 0)
          clonotype_sums_present = arrange(clonotype_sums_present, desc(Value)) ##### added sort based on total counts of clonotypes across cells
          rm(selected_assay)
          rownames(clonotype_sums_present)
        } 
      } else {
        character(0)  # Return empty if no data or assay not present
      }
    })
    
    selected_clonotypes <- reactiveVal(list())
    
    observeEvent(input$clonotypes, {
      current_selections <- selected_clonotypes()
      current_selections[[input$overlay_type]] <- input$clonotypes
      selected_clonotypes(current_selections)
    })
    
    output$clonotypeSelection <- renderUI({
      if (length(clonotypes()) == 0) {
        HTML("<p>No Clonotypes data available for the selected assay type.</p>")
      } else {
        tagList(
          selectInput(ns("clonotypes"), label = "Select clonotypes:",
                      choices = clonotypes(), multiple = TRUE),
          selectInput(ns("umapColorBy"), "Color by:", choices = metadataColumns(), selected = "seurat_clusters"),
          actionButton(ns("overlay_plot_action"), "Overlay Plot", style='color:#fff; background-color:#1c82e0')
        )
      }
    })
    
    output$resetbtn <- renderUI({
      if (length(clonotypes()) >  0) {
        actionButton(ns("resetPlot"), "Reset plot", style='color:#fff; background-color:#1c82e0')
      } 
    })
    
    umap_coords_with_clonotypes <- reactive({
      req(seuratObj())
      
      umap_coords_with_clonotypes <- data.frame()
      assays <- selected_clonotypes()
      
      for (assay in names(assays)) {
        clonotypes <- assays[[assay]]
        selected_assay <- seuratObj()[[assay]]$counts
        
        for (clonotype in clonotypes) {
          if (!is.null(selected_assay) && clonotype %in% rownames(selected_assay)) {
            if("umap" %in% Reductions(seuratObj())) {
              cells_with_clonotype <- which(selected_assay[clonotype, ] > 0)
              umap_coords <- data.frame(seuratObj()[["umap"]]@cell.embeddings)
              clonotype_coords <- umap_coords[rownames(umap_coords) %in% colnames(selected_assay)[cells_with_clonotype], ]
              
              clonotype_coords$clonotype <- clonotype
              clonotype_coords$assay <- assay
              clonotype_coords$clonotype <- as.factor(clonotype_coords$clonotype)
              umap_coords_with_clonotypes <- rbind(umap_coords_with_clonotypes, clonotype_coords)
            }
            if("tsne" %in% Reductions(seuratObj())) {
              cells_with_clonotype <- which(selected_assay[clonotype, ] > 0)
              tsne_coords <- data.frame(seuratObj()[["tsne"]]@cell.embeddings)
              clonotype_coords <- tsne_coords[rownames(tsne_coords) %in% colnames(selected_assay)[cells_with_clonotype], ]
              
              clonotype_coords$clonotype <- clonotype
              clonotype_coords$assay <- assay
              clonotype_coords$clonotype <- as.factor(clonotype_coords$clonotype)
              umap_coords_with_clonotypes <- rbind(umap_coords_with_clonotypes, clonotype_coords)
            }
          }
        }
      }
      rm(selected_assay)
      umap_coords_with_clonotypes
    })
    
    overlay_plot <- eventReactive(input$overlay_plot_action, {
      req(umap_coords_with_clonotypes(), seuratObj(), input$umapColorBy)
      
      # Check if either UMAP or tSNE reductions are present
      if (!"umap" %in% Reductions(seuratObj()) && !"tsne" %in% Reductions(seuratObj())) {
        shinyalert("Neither UMAP nor t-SNE reduction is available. Please run clustering Module.", type = "warning", confirmButtonCol = "#1c82e0")
        return()
      }
      
      insert_line_breaks <- function(text, width) {
        pattern <- paste0("(.{1,", width, "})(\\s|$|[^\\s])")
        wrapped_lines <- gsub(pattern, "\\1\n", text, perl = TRUE)
      }
  
      plot <- ggplotly(DimPlot(seuratObj(), pt.size = 1.5, label = F, label.size = 5, group.by = input$umapColorBy) +
                         RotatedAxis() +
                         guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) +
                         theme(aspect.ratio = 1, axis.text = element_text(colour = "black", size = 15, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white")))
      
      assay_colors <- c("TCRa" = "red", "TCRb" = "blue", "TCRdg" = "black", "BCRh" = "purple", "BCRl" = "green")
      umap_data <- umap_coords_with_clonotypes()

      for (clonotype in unique(umap_data$clonotype)) {
        assay <- unique(umap_data[umap_data$clonotype == clonotype, "assay"])[1]
        color <- assay_colors[[assay]]
        modified_clonotypes <- insert_line_breaks(clonotype, 20)
        
        if ("umap" %in% Reductions(seuratObj())) {
          plot <- plot %>% add_trace(
            data = umap_data[umap_data$clonotype == clonotype, ],
            x = ~umap_1,
            y = ~umap_2,
            type = "scatter",
            mode = "markers",
            marker = list(
              size = 7,
              symbol = "circle-open",
              color = color
            ),
            hoverinfo = "text",
            text = ~clonotype,
            name = modified_clonotypes,
            legendgroup = assay 
          )
        }
        if ("tsne" %in% Reductions(seuratObj())) {
          plot <- plot %>% add_trace(
            data = umap_data[umap_data$clonotype == clonotype, ],
            x = ~tSNE_1,
            y = ~tSNE_2,
            type = "scatter",
            mode = "markers",
            marker = list(
              size = 7,
              symbol = "circle-open",
              color = color
            ),
            hoverinfo = "text",
            text = ~clonotype,
            name = modified_clonotypes,  
            legendgroup = assay 
          )
        }
      }
      plot
    })
    
    static_overlay_plot <- reactive({
      req(umap_coords_with_clonotypes(), seuratObj(), input$clonotypes, input$umapColorBy)
      
      insert_line_breaks <- function(text, width) {
        pattern <- paste0("(.{1,", width, "})(\\s|$|[^\\s])")
        wrapped_lines <- gsub(pattern, "\\1\n", text, perl = TRUE)
      }
      
      p1 <- DimPlot(seuratObj(), pt.size = 1, label = FALSE, label.size = 5, group.by = input$umapColorBy) +
        RotatedAxis() + guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) +
        theme(aspect.ratio = 1, axis.text = element_text(colour = "black", size = 15, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white"))
      
      assay_colors <- c("TCRa" = "red", "TCRb" = "blue", "TCRdg" = "black", "BCRh" = "purple", "BCRl" = "green")
      umap_data <- umap_coords_with_clonotypes()
      
      for (clonotype in unique(umap_data$clonotype)) {
        if("umap" %in% Reductions(seuratObj())) {
          p1 <- p1 + geom_point(
            data = umap_data[umap_data$clonotype == clonotype, ],
            aes(x = umap_1, y = umap_2, shape = insert_line_breaks(clonotype, 20)),
            size = 1,
            fill = NA,
            color = assay_colors[umap_data[umap_data$clonotype == clonotype, "assay"]]) + 
            scale_shape_manual(values = rep(21, length(unique(umap_data$clonotype)))) + 
            guides(color = guide_legend(order=1),
                   shape = guide_legend(order=2, title = "")) +
            theme(legend.spacing.y = unit(-0.2, "cm"))
        }
        
        if("tsne" %in% Reductions(seuratObj())) {
          p1 <- p1 + geom_point(
            data = umap_data[umap_data$clonotype == clonotype, ],
            aes(x = tSNE_1, y = tSNE_2, shape = insert_line_breaks(clonotype, 20)),
            size = 1,
            fill = NA,
            color = assay_colors[umap_data[umap_data$clonotype == clonotype, "assay"]]) + 
            scale_shape_manual(values = rep(21, length(unique(umap_data$clonotype)))) +
            guides(shape = guide_legend(order=2, title = "")) +
            theme(legend.spacing.y = unit(-0.2, "cm"))
        }
      }
      rm(umap_data)
      p1
    })
    
    output$clonotypes_plot <- renderPlotly({
      req(overlay_plot())
      overlay_plot()
    })
    
    observe({
      if (!is.null(overlay_plot()) && !is.null(static_overlay_plot()) ) {
        shinyjs::show("ClonotypePlt")
      }
    })
    
    output$downloadPlot <- downloadHandler(
      filename = function() {
        paste0("CogentDS_ClonotypeOverlayPlot", "-", Sys.Date(), input$ClonotypePlt)
      },
      content = function(file) {
        req(static_overlay_plot())
        static_overlay_plot()
        ggsave(file, plot = static_overlay_plot(), width = 10, height = 10)
      },
      contentType = 'image'
    )
    
    observeEvent(input$resetPlot, {
      selected_clonotypes(list())
      updateSelectInput(session, ns("clonotypes"), selected = character(0))
    })
  })
}