#' Data Upload Server Component
#'  
#' This server module handles the server-side logic for data upload,
#' including reading the uploaded file, creating a Seurat object,
#' and storing it in a reactive value.
#'
#' @param id Unique identifier for the server module, ensuring isolated scope
#'           within the Shiny application.
#'
#' @return A reactive value containing the Seurat object created from
#'         the uploaded data.
#'         
#' Detailed Parameter Description:
#' - counts: A matrix of gene expression data where rows are genes and columns are cells/samples.
#' - project: Name of the project.
#' - min.cells: Minimum Number of cells in which a gene must be detected to be kept in the dataset.
#' - min.features: Minimum Number of features (genes) detected in a cell for it to be included in the analysis.
#'
#' @seealso \code{\link[Seurat]{CreateSeuratObject}}, \code{\link[Seurat]{PercentageFeatureSet}}
#' for details on Seurat object creation and estimation of % of all counts that belong to set of features. 
#'         
#' @export
#'
#' @examples
#' # In the server definition of a Shiny app
#' seuratObject <- dataUploadServer("dataUploadModule")
dataUploadServer <- function(id, value) {
  shiny::moduleServer(id, function(input, output, session) {
    ns <- session$ns
    data <- reactiveVal(NULL)
    
    observe({ 
      tryCatch({
        req(input$metadata_file)
        metadata = read.csv(input$metadata_file$datapath, header = TRUE, row.names = 1, check.names = F)
        updateSelectInput(session, "samples_col", choices = colnames(metadata))
      })
    })
    
    observe({ 
      tryCatch({
        req(input$metadata_file, input$samples_col)
        metadata = read.csv(input$metadata_file$datapath, header = TRUE, row.names = 1, check.names = F)
        sampletypes = unique(metadata[[input$samples_col]])
        sampletypes = c('Analyze all samples', sampletypes)
        updateSelectizeInput(session, "samples_pick", choices = sampletypes, selected = sampletypes[1], server = T)
      })
    })
    
    shiny::observeEvent(input$upload, {
      shinyjs::disable("upload")
      tryCatch({
        withProgress(message = 'Reading User inputs:', {
          incProgress(0.1, detail = 'tracking inputs and parameters')
          timeStart <- Sys.time()
          parameter_list = NULL
          if(input$fileoption == "Raw count matrix")
          {
            seuratObject = NULL
            req(input$file_raw, input$geneInfofile)
            seed_use = paste0('Seed Setting: set.seed(123456)')
            raw_counts_use = paste0('Raw count matrix: ', input$file_raw$name)
            geneinfo_use = paste0('Gene Info file: ', input$geneInfofile$name)
            metadata_use = paste0('Metadata file: ', input$metadata_file$name)
            projectName_use = paste0('Project Name: ', input$projectName)
            minCells_use = paste0('Minimum number of cells: ', input$minCells)
            minFeatures_use = paste0('Minimum number of features: ', input$minFeatures)
            parameter_list = c(seed_use, raw_counts_use, geneinfo_use, metadata_use, projectName_use, minCells_use, minFeatures_use)
            incProgress(0.1, detail = 'executing..')
            if((grepl(".csv$", input$file_raw$name)) | (grepl(".csv.gz$", input$file_raw$name)))
            {
              message('running check for number of cells')
              ch_check = fread(file = input$file_raw$datapath, sep = ",", data.table = F, verbose = T, check.names = FALSE) ## read the file using fread
              number_of_cells_in_data = length(colnames(ch_check)) - 1
              if(number_of_cells_in_data < 20) {
                seuratObject = NULL
                shinyalert("Too few cells in a single cell RNA experiment", "CogentDS v2.0 expects >= 20 cells", type = "warning", confirmButtonCol = "#1c82e0")
                stop ()
              }
              if(input$metadata_question == "Yes")
              {
                req(input$metadata_file)
                metadata = read.csv(input$metadata_file$datapath, header = TRUE, row.names = 1, check.names = F)
                message("read metadata")
                if(!is.null(input$geneInfofile)) {
                  if(is.null(input$samples_pick))
                  {
                    seuratObject = NULL
                    seuratObject_out = NULL
                    shinyalert("Sample Selection not done", "Please select atleast one sample", type = "error", confirmButtonCol = "#1c82e0")
                    stop()
                  }
                  if(length(input$samples_pick) > 1)
                  {
                    if('Analyze all samples' %in% input$samples_pick) {
                      source('R/Setup_Funtions/bin/FastRead_largeCSV.R', local = TRUE)
                      seuratObject = seuratObject_out
                      message("object ready")
                    } else {
                      source('R/Setup_Funtions/bin/FastRead_largeCSV.R', local = TRUE)
                      seuratObject = seuratObject_out
                      Idents(seuratObject) = input$samples_col
                      seuratObject$orig.ident = seuratObject[[input$samples_col]]
                      seuratObject = subset(seuratObject, idents = input$samples_pick)
                      message("object ready")
                    }
                  }
                  if(length(input$samples_pick) == 1)
                  {
                    if(input$samples_pick == 'Analyze all samples') {
                      source('R/Setup_Funtions/bin/FastRead_largeCSV.R', local = TRUE)
                      seuratObject = seuratObject_out
                      message("object ready")
                    } else {
                      source('R/Setup_Funtions/bin/FastRead_largeCSV.R', local = TRUE)
                      seuratObject = seuratObject_out
                      Idents(seuratObject) = input$samples_col
                      seuratObject$orig.ident = seuratObject[[input$samples_col]]
                      seuratObject = subset(seuratObject, idents = input$samples_pick)
                      message("object ready")
                    }
                  }
                } else {
                  shinyalert('Gene Info file is missing or set to NULL', 'Please refresh app and upload Gene Info file', type = "warning", confirmButtonCol = "#1c82e0")
                  stop ()
                }
              }
              if(input$metadata_question == "No")
              {
                metadata = NULL
                if(!is.null(input$geneInfofile)) {
                  source('R/Setup_Funtions/bin/FastRead_largeCSV.R', local = TRUE)
                  seuratObject = seuratObject_out
                  message("object ready")
                } else {
                  seuratObject = NULL
                  seuratObject_out = NULL
                  shinyalert('Gene Info file is missing or set to NULL', 'Please refresh app and upload Gene Info file', type = "warning", confirmButtonCol = "#1c82e0")
                }
              }
              incProgress(0.3, detail = 'seurat object creation completed')
              rm(seuratObject_out)
            }
          }
          if(input$fileoption == "Processed Data from CogentAP")
          {
            seuratObject = NULL
            junction_reads <- NULL
            span_frags <- NULL
            TCRb <- NULL
            TCRa <- NULL
            TCRdg <- NULL
            BCRh <- NULL
            BCRl <- NULL
            df_trans <- NULL
            df_temp <- NULL
            req(input$file, input$RNA_use)
            seed_use = paste0('Seed Setting: set.seed(123456)')
            cogentAP_use = paste0('CogentAP processed data: ', input$file$name)
            metadata_use = paste0('Metadata file: ', input$metadata_file$name)
            projectName_use = paste0('Project Name: ', input$projectName)
            minCells_use = paste0('Minimum number of cells: ', input$minCells)
            minFeatures_use = paste0('Minimum number of features: ', input$minFeatures)
            parameter_list = c(seed_use, cogentAP_use, metadata_use, projectName_use, minCells_use, minFeatures_use)
            incProgress(0.1, detail = 'executing...')
            if (grepl(".rda$", input$file$name)) {
              load(input$file$datapath)
              df <- get("CogentDS_data")[["raw_data"]]$gm
              
              number_of_cells_in_data = length(colnames(df))
              if(number_of_cells_in_data < 20) {
                seuratObject = NULL
                shinyalert("Too few cells in a single cell RNA experiment", "CogentDS v2.0 expects >= 20 cells", type = "warning", confirmButtonCol = "#1c82e0")
                stop ()
              }
              
              metadata_import <-  CogentDS_data[["raw_data"]][["metadata"]]
              # Get fusion genes
              # Check if "overlay_raw_data" exists in the loaded data
              if ("overlay_raw_data" %in% names(get("CogentDS_data"))) {
                sce <- get("CogentDS_data")[["overlay_raw_data"]]
                
                if ("gene_fusion" %in% altExpNames(sce)) {
                  alt_exp <- altExp(sce, "gene_fusion")
                  junction_reads <- assay(alt_exp, "junction_reads")
                  span_frags <- assay(alt_exp, "span_frags")
                }
                
                # Get clonotypes data
                if (any(c("TCRb", "TCRa", "BCRh", "BCRl") %in% altExpNames(sce))) {
                  if ("TCRb" %in% altExpNames(sce)) TCRb <- assay(altExp(sce, "TCRb"))
                  if ("TCRa" %in% altExpNames(sce)) TCRa <- assay(altExp(sce, "TCRa"))
                  if ("BCRh" %in% altExpNames(sce)) BCRh <- assay(altExp(sce, "BCRh"))
                  if ("BCRl" %in% altExpNames(sce)) BCRl <- assay(altExp(sce, "BCRl"))
                }
              }
              seuratObject <- Seurat::CreateSeuratObject(counts = df, project = input$projectName, min.cells = input$minCells, min.features = input$minFeatures)
              if(input$metadata_question == "Yes") {
                req(input$metadata_file)
                metadata = read.csv(input$metadata_file$datapath, header = TRUE, row.names = 1)
                check_cells = intersect(rownames(metadata), Cells(seuratObject))
                if(length(check_cells) == length(Cells(seuratObject))) {
                  message("read metadata")
                  if(length(input$samples_pick) == 1) {
                    if(input$samples_pick == 'Analyze all samples') {
                      if(length(colnames(metadata)) == 1) {
                        metadata = metadata[Cells(seuratObject),]
                        seuratObject = AddMetaData(seuratObject, metadata = metadata, col.name = input$samples_col)
                      } else {
                        metadata = metadata[Cells(seuratObject),]
                        seuratObject = AddMetaData(seuratObject, metadata = metadata)
                      }
                    }
                    else {
                      if(length(colnames(metadata)) == 1) {
                        metadata = metadata[Cells(seuratObject),]
                        seuratObject = AddMetaData(seuratObject, metadata = metadata, col.name = input$samples_col)
                      } else {
                        metadata = metadata[Cells(seuratObject),]
                        seuratObject = AddMetaData(seuratObject, metadata = metadata)
                      }
                      Idents(seuratObject) = input$samples_col
                      seuratObject$orig.ident = seuratObject[[input$samples_col]]
                      seuratObject = subset(seuratObject, idents = input$samples_pick)
                    }
                  }
                  if(length(input$samples_pick) > 1) {
                    if('Analyze all samples' %in% input$samples_pick) {
                      if(length(colnames(metadata)) == 1) {
                        message("object ready")
                        metadata = metadata[Cells(seuratObject),]
                        seuratObject = AddMetaData(seuratObject, metadata = metadata, col.name = input$samples_col)
                      } else {
                        message("object ready")
                        metadata = metadata[Cells(seuratObject),]
                        seuratObject = AddMetaData(seuratObject, metadata = metadata)
                      }
                    } else {
                      if(length(colnames(metadata)) == 1) {
                        message("object ready")
                        metadata = metadata[Cells(seuratObject),]
                        seuratObject = AddMetaData(seuratObject, metadata = metadata, col.name = input$samples_col)
                      } else {
                        message("object ready")
                        metadata = metadata[Cells(seuratObject),]
                        seuratObject = AddMetaData(seuratObject, metadata = metadata)
                      }
                      Idents(seuratObject) = input$samples_col
                      seuratObject$orig.ident = seuratObject[[input$samples_col]]
                      seuratObject = subset(seuratObject, idents = input$samples_pick)
                    }
                  }
                } else {
                  message ("Fatal Error: Cells in metadata do not match user provided seurat object; Metadata should have cells in the same order and equal length as in user provided seurat object")
                }
              }
              if(input$metadata_question == "No") {
                metadata = NULL
                seuratObject <- Seurat::CreateSeuratObject(counts = df, project = input$projectName, min.cells = input$minCells, min.features = input$minFeatures, meta.data = metadata_import)
              }
              rm(df)
            }
            if (grepl(".rds$", input$file$name)) {       ### added the .rds acceptance
              df_temp <- readRDS(input$file$datapath)
              ###### check number of cells ########
              number_of_cells_in_data = length(Cells(df_temp))
              if(number_of_cells_in_data < 20) {
                shinyalert("Too few cells in a single cell RNA experiment", "CogentDS v2.0 expects >= 20 cells", type = "warning", confirmButtonCol = "#1c82e0")
                stop ()
              }

              if(("RNA" %in% names(df_temp)) && ("Exon_RNA" %in% names(df_temp))) {
                if(input$RNA_use == "Exon+Intron") {
                  message('Analyzing Exon+Intron expression of genes for downstream analysis')
                  df <- df_temp[["RNA"]]$counts
                  df_genes_mean_exp = data.frame(rowMeans(df))
                  colnames(df_genes_mean_exp) = "counts"
                  df_genes_mean_exp_expressed = df_genes_mean_exp %>% dplyr::filter(counts > 0)
                  df = df[rownames(df_genes_mean_exp_expressed),]
                  expression_profileanalyzed = paste0('CogentDS analyzed Exon+Intron expression profile')
                  parameter_list = c(parameter_list, expression_profileanalyzed)
                }
                if(input$RNA_use == "Exon-Only") {
                  message('Analyzing Exon-Only expression of genes for downstream analysis')
                  df <- df_temp[["Exon_RNA"]]$counts
                  df_genes_mean_exp = data.frame(rowMeans(df))
                  colnames(df_genes_mean_exp) = "counts"
                  df_genes_mean_exp_expressed = df_genes_mean_exp %>% dplyr::filter(counts > 0)
                  df = df[rownames(df_genes_mean_exp_expressed),]
                  expression_profileanalyzed = paste0('CogentDS analyzed Exon-Only expression profile')
                  parameter_list = c(parameter_list, expression_profileanalyzed)
                }
              } else {
                message('Analyzing RNA expression profile; CogentDS suspects uploaded data is not processed using CogentAP v3.1; raising info alert')
                df <- df_temp[["RNA"]]$counts
                df_genes_mean_exp = data.frame(rowMeans(df))
                colnames(df_genes_mean_exp) = "counts"
                df_genes_mean_exp_expressed = df_genes_mean_exp %>% dplyr::filter(counts > 0)
                df = df[rownames(df_genes_mean_exp_expressed),]
                expression_profileanalyzed = paste0('CogentDS analyzed RNA expression profile; warning: CogentDS suspects data uploaded is not processed using CogentAP v3.1')
                parameter_list = c(parameter_list, expression_profileanalyzed)
                shinyalert("Uploaded data is not processed using CogentAP v3.1", "Please See: Downstream analysis will be based on RNA assay", type = "info", confirmButtonCol = "#1c82e0")
              }    
              
              metadata_import = data.frame(df_temp@meta.data)
              cols_not_needed_import = c('orig.ident', 'nCount_RNA', 'nFeature_RNA', 'seurat_clusters', 'RNA_snn_res.0.5', 'RNA_snn_res.0.8', 'RNA_snn_res.1')
              find_not_needed_import = which(colnames(metadata_import) %in% cols_not_needed_import)
              
              if(length(find_not_needed_import) > 0) {
                metadata_import = metadata_import %>% dplyr::select(!colnames(metadata_import)[find_not_needed_import])
                message('removed unwanted colums')
              }
              
              if (any(c("junction_reads", "span_frags") %in%  names(df_temp))){
                if("junction_reads" %in% names(df_temp)) {
                  junction_reads <- df_temp[["junction_reads"]]$counts
                }
                if("span_frags" %in% names(df_temp)) {
                  span_frags <- df_temp[["span_frags"]]$counts
                }
                message("read fusion")
              }
              
              if (any(c("BCRl", "BCRh", "TCRa", "TCRb", "TCRdg") %in%  names(df_temp))){
                if("BCRh" %in% names(df_temp)) {
                  BCRh <- df_temp[["BCRh"]]$counts
                }
                if("BCRl" %in% names(df_temp)) {
                  BCRl <- df_temp[["BCRl"]]$counts
                }
                if("TCRa" %in% names(df_temp)) {
                  TCRa <- df_temp[["TCRa"]]$counts
                }
                if("TCRb" %in% names(df_temp)) {
                  TCRb <- df_temp[["TCRb"]]$counts
                }
                if("TCRdg" %in% names(df_temp)) {
                  TCRdg <- df_temp[["TCRdg"]]$counts
                }
                message("read clonotype")
              }
              
              if(any('Transcript_RNA' %in%  names(df_temp)))
              {
                df_trans <- df_temp[["Transcript_RNA"]]$counts
                df_trans_mean_exp = data.frame(rowMeans(df_trans))
                colnames(df_trans_mean_exp) = "counts"
                df_trans_mean_exp_expressed = df_trans_mean_exp %>% dplyr::filter(counts > 0)
                df_trans = df_trans[rownames(df_trans_mean_exp_expressed),]
                time = Sys.time()
                time = gsub(":", "-", time)
                time = gsub(" ", "-", time)
                incProgress(0.1, detail = 'Saving isoform to disk for later access..')
                save_isoform_data <- paste0('R/Setup_Funtions/bin/temp/', 'isoform_data_', time, "_file.rds")
                sc_isoform = CreateSeuratObject(df_trans)
                saveRDS(sc_isoform, file = save_isoform_data)
                sc_isoform_filename = paste0('Isoform Data: ', save_isoform_data)
                parameter_list = c(parameter_list, sc_isoform_filename)
                message("read isoform and saved to temp folder for later access")
                rm(df_trans)
              }
              if(input$metadata_question == "Yes") {
                seuratObject <- CreateSeuratObject(counts = df, project = input$projectName, min.cells = input$minCells, min.features = input$minFeatures, meta.data = metadata_import)
                req(input$metadata_file)
                metadata = read.csv(input$metadata_file$datapath, header = TRUE, row.names = 1)
                check_cells = intersect(rownames(metadata), Cells(seuratObject))
                if(length(check_cells) == length(Cells(seuratObject))) {
                  message("read metadata")
                  if(length(input$samples_pick) == 1) {
                    if(input$samples_pick == 'Analyze all samples') {
                      if(length(colnames(metadata)) == 1) {
                        metadata = metadata[Cells(seuratObject),]
                        seuratObject = AddMetaData(seuratObject, metadata = metadata, col.name = input$samples_col)
                      } else {
                        metadata = metadata[Cells(seuratObject),]
                        seuratObject = AddMetaData(seuratObject, metadata = metadata)
                      }
                    }
                    else {
                      if(length(colnames(metadata)) == 1) {
                        metadata = metadata[Cells(seuratObject),]
                        seuratObject = AddMetaData(seuratObject, metadata = metadata, col.name = input$samples_col)
                      } else {
                        metadata = metadata[Cells(seuratObject),]
                        seuratObject = AddMetaData(seuratObject, metadata = metadata)
                      }
                      Idents(seuratObject) = input$samples_col
                      seuratObject$orig.ident = seuratObject[[input$samples_col]]
                      seuratObject = subset(seuratObject, idents = input$samples_pick)
                    }
                  }
                  if(length(input$samples_pick) > 1) {
                    if('Analyze all samples' %in% input$samples_pick) {
                      if(length(colnames(metadata)) == 1) {
                        message("object ready")
                        metadata = metadata[Cells(seuratObject),]
                        seuratObject = AddMetaData(seuratObject, metadata = metadata, col.name = input$samples_col)
                      } else {
                        message("object ready")
                        metadata = metadata[Cells(seuratObject),]
                        seuratObject = AddMetaData(seuratObject, metadata = metadata)
                      }
                    } else {
                      if(length(colnames(metadata)) == 1) {
                        message("object ready")
                        metadata = metadata[Cells(seuratObject),]
                        seuratObject = AddMetaData(seuratObject, metadata = metadata, col.name = input$samples_col)
                      } else {
                        message("object ready")
                        metadata = metadata[Cells(seuratObject),]
                        seuratObject = AddMetaData(seuratObject, metadata = metadata)
                      }
                      Idents(seuratObject) = input$samples_col
                      seuratObject$orig.ident = seuratObject[[input$samples_col]]
                      seuratObject = subset(seuratObject, idents = input$samples_pick)
                    }
                  }
                } else {
                  message ("Fatal Error: Cells in metadata do not match user provided seurat object; Metadata should have cells in the same order and equal length as in user provided seurat object")
                }
              }
              if(input$metadata_question == "No") {
                metadata = NULL
                seuratObject <- CreateSeuratObject(counts = df, project = input$projectName, min.cells = input$minCells, min.features = input$minFeatures, meta.data = metadata_import)
              }
              rm(df)
            }
            if(!is.null(seuratObject)) {
              if ((!is.null(junction_reads)) && (nrow(junction_reads) > 0)) {
                intersect_col = intersect(colnames(junction_reads), Cells(seuratObject))
                if(length(intersect_col) > 0) {
                  junction_reads = junction_reads[,intersect_col]
                  seuratObject[["junction_reads"]] = CreateAssay5Object(counts = junction_reads)
                  message("junction ready")
                  rm(junction_reads)
                }
              }
              if ((!is.null(span_frags)) && (nrow(span_frags) > 0)) {
                intersect_col = intersect(colnames(span_frags), Cells(seuratObject))
                if(length(intersect_col) > 0) {
                  span_frags = span_frags[,intersect_col]
                  seuratObject[["span_frags"]] = CreateAssay5Object(counts = span_frags)
                  message("span ready")
                  rm(span_frags)
                }
              }
              if ((!is.null(BCRh)) && (nrow(BCRh) > 0)) {
                intersect_col = intersect(colnames(BCRh), Cells(seuratObject))
                if(length(intersect_col) > 0) {
                  BCRh = BCRh[,intersect_col]
                  seuratObject[["BCRh"]] = CreateAssay5Object(counts = BCRh)
                  message("BCRh ready")
                  rm(BCRh)
                }
              } 
              if ((!is.null(BCRl)) && (nrow(BCRl) > 0)) {
                intersect_col = intersect(colnames(BCRl), Cells(seuratObject))
                if(length(intersect_col) > 0) {
                  BCRl = BCRl[,intersect_col]
                  seuratObject[["BCRl"]] = CreateAssay5Object(counts = BCRl)
                  message("BCRl ready")
                  rm(BCRl)
                }
              }
              if ((!is.null(TCRa)) && (nrow(TCRa) > 0)) {
                intersect_col = intersect(colnames(TCRa), Cells(seuratObject))
                if(length(intersect_col) > 0) {
                  TCRa = TCRa[,intersect_col]
                  seuratObject[["TCRa"]] = CreateAssay5Object(counts = TCRa)
                  message("TCRa ready")
                  rm(TCRa)
                }
              }
              if ((!is.null(TCRb)) && (nrow(TCRb) > 0)) {
                intersect_col = intersect(colnames(TCRb), Cells(seuratObject))
                if(length(intersect_col) > 0) {
                  TCRb = TCRb[,intersect_col]
                  seuratObject[["TCRb"]] = CreateAssay5Object(counts = TCRb)
                  message("TCRb ready")
                  rm(TCRb)
                }
              }
              if ((!is.null(TCRdg)) && (nrow(TCRdg) > 0)) {
                intersect_col = intersect(colnames(TCRdg), Cells(seuratObject))
                if(length(intersect_col) > 0) {
                  TCRdg = TCRdg[,intersect_col]
                  seuratObject[["TCRdg"]] = CreateAssay5Object(counts = TCRdg)
                  message("TCRdg ready")
                  rm(TCRdg)
                }
              }

              if(("Gene_Biotype" %in% colnames(df_temp[['RNA']][[]])) && ("Gene_Length" %in% colnames(df_temp[['RNA']][[]]))) {
                biotype_info = cbind(Features(df_temp), df_temp[['RNA']][[]]$Gene_Biotype, df_temp[['RNA']][[]]$Gene_Length)
                biotype_info = data.frame(biotype_info)
                colnames(biotype_info) = c("Gene_Name", "Gene_Biotype", "Gene_Length")
                rownames(biotype_info) = biotype_info$Gene_Name
                biotype_info = biotype_info[Features(seuratObject),]
                seuratObject[['RNA']][[]]$Gene_Biotype = biotype_info$Gene_Biotype
                seuratObject[['RNA']][[]]$Gene_Length = biotype_info$Gene_Length
                message('added biotype and length info for features')
                incProgress(0.1, detail = 'Saving biotype plot data for later access..')
                avg_expression = AggregateExpression(seuratObject, assays = "RNA", group.by = "orig.ident")
                avg_expression = data.frame(avg_expression)
                avg_expression$RNA_Biotype = seuratObject[['RNA']][[]]$Gene_Biotype
                avg_expression_plot = reshape2::melt(avg_expression, id.vars = "RNA_Biotype", value.name = "RNA")
                time = Sys.time()
                time = gsub(":", "-", time)
                time = gsub(" ", "-", time)
                save_biotype_data <- paste0('R/Setup_Funtions/bin/temp/', 'biotype_data_', time, "_file.txt")
                write.table(avg_expression_plot, save_biotype_data, quote = F, sep = "\t")
                sc_biotype_filename = paste0('RNABiotype Data: ', save_biotype_data)
                parameter_list = c(parameter_list, sc_biotype_filename)
                rm(avg_expression_plot)
                rm(avg_expression)
              } else {
                message('biotype and length info for features not found....')
              }
              
              incProgress(0.2, detail = 'seurat object creation completed')
            } else {
              message ("Fatal: Seurat object creation failed; Please check your inputs")
              stop("Error: Seurat object creation failed: Fatal")
            }
            rm(df_temp)
            gc()
          }
          if(input$fileoption == "Example data") {
            seed_use = paste0('Seed Setting: set.seed(123456)')
            df_temp <- readRDS('data/exampleData_PBMC_2k_with_Isoform-Fusion-Clonotype.rds')

            if(("RNA" %in% names(df_temp)) && ("Exon_RNA" %in% names(df_temp))) {
                if(input$RNA_use == "Exon+Intron") {
                  message('Analyzing Exon+Intron expression of genes for downstream analysis')
                  df <- df_temp[["RNA"]]$counts
                  df_genes_mean_exp = data.frame(rowMeans(df))
                  colnames(df_genes_mean_exp) = "counts"
                  df_genes_mean_exp_expressed = df_genes_mean_exp %>% dplyr::filter(counts > 0)
                  df = df[rownames(df_genes_mean_exp_expressed),]
                  expression_profileanalyzed = paste0('CogentDS analyzed Exon+Intron expression profile')
                }
                if(input$RNA_use == "Exon-Only") {
                  message('Analyzing Exon-Only expression of genes for downstream analysis')
                  df <- df_temp[["Exon_RNA"]]$counts
                  df_genes_mean_exp = data.frame(rowMeans(df))
                  colnames(df_genes_mean_exp) = "counts"
                  df_genes_mean_exp_expressed = df_genes_mean_exp %>% dplyr::filter(counts > 0)
                  df = df[rownames(df_genes_mean_exp_expressed),]
                  expression_profileanalyzed = paste0('CogentDS analyzed Exon-Only expression profile')
                }
            } else {
              message('Analyzing RNA expression profile; CogentDS suspects uploaded data is not processed using CogentAP v3.1; raising info alert')
              df <- df_temp[["RNA"]]$counts
              df_genes_mean_exp = data.frame(rowMeans(df))
              colnames(df_genes_mean_exp) = "counts"
              df_genes_mean_exp_expressed = df_genes_mean_exp %>% dplyr::filter(counts > 0)
              df = df[rownames(df_genes_mean_exp_expressed),]
              expression_profileanalyzed = paste0('CogentDS analyzed RNA expression profile; warning: CogentDS suspects data uploaded is not processed using CogentAP v3.1')
            }
              
            if (any(c("junction_reads", "span_frags") %in%  names(df_temp))){
              junction_reads <- df_temp[["junction_reads"]]$counts
              span_frags <- df_temp[["span_frags"]]$counts
              message("read fusion")
            }
            
            if (any(c("BCRl", "BCRh", "TCRa", "TCRb", "TCRdg") %in%  names(df_temp))){
              BCRh <- df_temp[["BCRh"]]$counts
              BCRl <- df_temp[["BCRl"]]$counts
              TCRa <- df_temp[["TCRa"]]$counts
              TCRb <- df_temp[["TCRb"]]$counts
              TCRdg <- df_temp[["TCRdg"]]$counts
              message("read clonotype")
            }
              
            if(any('Transcript_RNA' %in%  names(df_temp)))
            {
              df_trans <- df_temp[["Transcript_RNA"]]$counts
              df_trans_mean_exp = data.frame(rowMeans(df_trans))
              colnames(df_trans_mean_exp) = "counts"
              df_trans_mean_exp_expressed = df_trans_mean_exp %>% dplyr::filter(counts > 0)
              df_trans = df_trans[rownames(df_trans_mean_exp_expressed),]
              time = Sys.time()
              time = gsub(":", "-", time)
              time = gsub(" ", "-", time)
              incProgress(0.1, detail = 'Saving isoform to disk for later access..')
              save_isoform_data <- paste0('R/Setup_Funtions/bin/temp/', 'isoform_data_', time, "_file.rds")
              sc_isoform = CreateSeuratObject(df_trans)
              saveRDS(sc_isoform, file = save_isoform_data)
              sc_isoform_filename = paste0('Isoform Data: ', save_isoform_data)
              message("read isoform and saved to temp folder for later access")
              rm(df_trans)
            }

            seuratObject <- CreateSeuratObject(counts = df, meta.data = df_temp@meta.data)

            rm(df)

            if(!is.null(seuratObject)) {
              if ((!is.null(junction_reads)) && (nrow(junction_reads) > 0)) {
                intersect_col = intersect(colnames(junction_reads), Cells(seuratObject))
                if(length(intersect_col) > 0) {
                  junction_reads = junction_reads[,intersect_col]
                  seuratObject[["junction_reads"]] = CreateAssay5Object(counts = junction_reads)
                  message("junction ready")
                  rm(junction_reads)
                }
              }
              if ((!is.null(span_frags)) && (nrow(span_frags) > 0)) {
                intersect_col = intersect(colnames(span_frags), Cells(seuratObject))
                if(length(intersect_col) > 0) {
                  span_frags = span_frags[,intersect_col]
                  seuratObject[["span_frags"]] = CreateAssay5Object(counts = span_frags)
                  message("span ready")
                  rm(span_frags)
                }
              }
              if ((!is.null(BCRh)) && (nrow(BCRh) > 0)) {
                intersect_col = intersect(colnames(BCRh), Cells(seuratObject))
                if(length(intersect_col) > 0) {
                  BCRh = BCRh[,intersect_col]
                  seuratObject[["BCRh"]] = CreateAssay5Object(counts = BCRh)
                  message("BCRh ready")
                  rm(BCRh)
                }
              } 
              if ((!is.null(BCRl)) && (nrow(BCRl) > 0)) {
                intersect_col = intersect(colnames(BCRl), Cells(seuratObject))
                if(length(intersect_col) > 0) {
                  BCRl = BCRl[,intersect_col]
                  seuratObject[["BCRl"]] = CreateAssay5Object(counts = BCRl)
                  message("BCRl ready")
                  rm(BCRl)
                }
              }
              if ((!is.null(TCRa)) && (nrow(TCRa) > 0)) {
                intersect_col = intersect(colnames(TCRa), Cells(seuratObject))
                if(length(intersect_col) > 0) {
                  TCRa = TCRa[,intersect_col]
                  seuratObject[["TCRa"]] = CreateAssay5Object(counts = TCRa)
                  message("TCRa ready")
                  rm(TCRa)
                }
              }
              if ((!is.null(TCRb)) && (nrow(TCRb) > 0)) {
                intersect_col = intersect(colnames(TCRb), Cells(seuratObject))
                if(length(intersect_col) > 0) {
                  TCRb = TCRb[,intersect_col]
                  seuratObject[["TCRb"]] = CreateAssay5Object(counts = TCRb)
                  message("TCRb ready")
                  rm(TCRb)
                }
              }
              if ((!is.null(TCRdg)) && (nrow(TCRdg) > 0)) {
                intersect_col = intersect(colnames(TCRdg), Cells(seuratObject))
                if(length(intersect_col) > 0) {
                  TCRdg = TCRdg[,intersect_col]
                  seuratObject[["TCRdg"]] = CreateAssay5Object(counts = TCRdg)
                  message("TCRdg ready")
                  rm(TCRdg)
                }
              }

              if(("Gene_Biotype" %in% colnames(df_temp[['RNA']][[]])) && ("Gene_Length" %in% colnames(df_temp[['RNA']][[]]))) {
                biotype_info = cbind(Features(df_temp), df_temp[['RNA']][[]]$Gene_Biotype, df_temp[['RNA']][[]]$Gene_Length)
                biotype_info = data.frame(biotype_info)
                colnames(biotype_info) = c("Gene_Name", "Gene_Biotype", "Gene_Length")
                rownames(biotype_info) = biotype_info$Gene_Name
                biotype_info = biotype_info[Features(seuratObject),]
                seuratObject[['RNA']][[]]$Gene_Biotype = biotype_info$Gene_Biotype
                seuratObject[['RNA']][[]]$Gene_Length = biotype_info$Gene_Length
                message('added biotype and length info for features')
                incProgress(0.1, detail = 'saving biotype plot data for later access')
                avg_expression = AggregateExpression(seuratObject, assays = "RNA", group.by = "orig.ident")
                avg_expression = data.frame(avg_expression)
                avg_expression$RNA_Biotype = seuratObject[['RNA']][[]]$Gene_Biotype
                avg_expression_plot = reshape2::melt(avg_expression, id.vars = "RNA_Biotype", value.name = "RNA")
                time = Sys.time()
                time = gsub(":", "-", time)
                time = gsub(" ", "-", time)
                save_biotype_data <- paste0('R/Setup_Funtions/bin/temp/', 'biotype_data_', time, "_file.txt")
                write.table(avg_expression_plot, save_biotype_data, quote = F, sep = "\t")
                sc_biotype_filename = paste0('RNABiotype Data: ', save_biotype_data)
                rm(avg_expression_plot)
                rm(avg_expression)
              } else {
                message('biotype and length info for features not found....')
              }
            }
            rm(df_temp)
            examplefile = paste0('Example data: ', 'exampleData_PBMC_2k_with_Isoform-Fusion-Clonotype.rds')
            projectName_use = paste0('Project Name: ', 'SeuratProject')
            minCells_use = paste0('Minimum number of cells: ', 0)
            minFeatures_use = paste0('Minimum number of features: ', 0)
            incProgress(0.1, detail = 'executing...')
            sample_not_needed = 'Sample'
            find_not_needed_sample = which(colnames(seuratObject@meta.data) %in% sample_not_needed)
            
            if(length(find_not_needed_sample) > 0) {
              seuratObject@meta.data = seuratObject@meta.data %>% dplyr::select(!colnames(seuratObject@meta.data)[find_not_needed_sample])
              message('Removed unwanted columns from the metadata')
            }
            parameter_list = c(seed_use, examplefile, projectName_use, minCells_use, minFeatures_use, expression_profileanalyzed, sc_isoform_filename, sc_biotype_filename)
            incProgress(0.2, detail = 'seurat object creation completed')
            message("object ready") 
          }
          
          if (!is.null(seuratObject)) {
            incProgress(0.1, detail = 'seurat object checks in progress')
            seuratObject@assays$user_parameter_tracker = parameter_list  ######## created a tracker for user args
            
            number_of_cells_in_data = length(Cells(seuratObject))
            if(number_of_cells_in_data < 20) {
              shinyalert("Too few cells; < 20 found, please adjust the filtering parameters", "CogentDS v2.0 expects >= 20 cells", type = "warning", confirmButtonCol = "#1c82e0")
              stop ()
            }
            
            ###### add dry run paramerters
            dryrun1 = paste0('dry run paramerters; normalization method: LogNormalize')
            dryrun2 = paste0('dry run paramerters; nfeatures: 2000')
            dryrun3 = paste0('dry run paramerters; feature selection method: vst')
            dryrun4 = paste0('dry run paramerters; npcs: 10')
            dryrun5 = paste0('dry run paramerters; number of dimensions: 10')
            dryrun6 = paste0('dry run paramerters; clustering resolution: 0.5')
            parameter_list = c(parameter_list, dryrun1, dryrun2, dryrun3, dryrun4, dryrun5, dryrun6)
            seuratObject@assays$user_parameter_tracker = parameter_list
            
            message("parameter tracking is on")
            message("Object ready for stats!")
            
            # Estimating the percentage of mitochondrial genes as an initial quality control metric
            ##seuratObject[["percent.mt"]] <- Seurat::PercentageFeatureSet(seuratObject, pattern = "^MT-")
            check_Mito_presence_human = grep("^MT-", rownames(seuratObject), value = TRUE, ignore.case = FALSE)
            check_Mito_presence_mouse = grep("^mt-", rownames(seuratObject), value = TRUE, ignore.case = FALSE)
            if(length(check_Mito_presence_human) > 0) {
              seuratObject[["percent.mt"]] <- Seurat::PercentageFeatureSet(seuratObject, pattern = "^MT-")
            }
            if(length(check_Mito_presence_mouse) > 0) {
              seuratObject[["percent.mt"]] <- Seurat::PercentageFeatureSet(seuratObject, pattern = "^mt-")
            }
            if((length(check_Mito_presence_human) == 0) && (length(check_Mito_presence_mouse) == 0)) {
              seuratObject[["percent.mt"]] = rep(0, ncol(seuratObject))
            }
            metadata = data.frame(seuratObject@meta.data)
            cols_imp = c('Barcoded_Reads', 'Trimmed_Reads', 'Unmapped_Reads', 'Mapped_Reads', 'Uniquely_Mapped_Reads', 'Multimapped_Reads', 'Chimeric_Reads', 'No_of_Reads_w_UMIs', 'Reads_After_Dedup_nUMIs_USSs', 'Exon_nUMIs_USSs', 'Intron_nUMIs_USSs', 'Gene_nUMIs_USSs', 'Intergenic_Reads', 'Ribosomal_Reads', 'Mitochondrial_Reads', 'Exon_Reads', 'Intron_Reads', 'Gene_Reads')
            find_imp = which(colnames(metadata) %in% cols_imp)
            
            if(length(find_imp) > 0) {
              if('Intergenic_Reads' %in% colnames(metadata)) {
                seuratObject$IntergenicReadFraction = as.numeric(seuratObject$Intergenic_Reads)/as.numeric(seuratObject$Mapped_Reads)
                seuratObject$Intergenic_reads_percentage = (as.numeric(seuratObject$Intergenic_Reads)/as.numeric(seuratObject$Mapped_Reads))*100
              } else {
                seuratObject$IntergenicReadFraction = rep(0, nrow(seuratObject@meta.data))
                seuratObject$Intergenic_reads_percentage = rep(0, nrow(seuratObject@meta.data))
                
              }
              if('Ribosomal_Reads' %in% colnames(metadata)) {
                sumna = sum(is.na(seuratObject@meta.data$Ribosomal_Reads))
                if(sumna > 0) {
                  seuratObject$RibosomalReadFraction = rep(0, nrow(seuratObject@meta.data))
                  seuratObject$Ribosomal_reads_percentage = rep(0, nrow(seuratObject@meta.data))
                } else {
                  seuratObject$RibosomalReadFraction = as.numeric(seuratObject$Ribosomal_Reads)/as.numeric(seuratObject$Trimmed_Reads)
                  seuratObject$Ribosomal_reads_percentage = (as.numeric(seuratObject$Ribosomal_Reads)/as.numeric(seuratObject$Trimmed_Reads))*100
                } 
              } else {
                seuratObject$RibosomalReadFraction = rep(0, nrow(seuratObject@meta.data))
                seuratObject$Ribosomal_reads_percentage = rep(0, nrow(seuratObject@meta.data))
              }
            } else {
              seuratObject$IntergenicReadFraction = rep(0, nrow(seuratObject@meta.data))
              seuratObject$RibosomalReadFraction = rep(0, nrow(seuratObject@meta.data))
              seuratObject$Intergenic_reads_percentage = rep(0, nrow(seuratObject@meta.data))
              seuratObject$Ribosomal_reads_percentage = rep(0, nrow(seuratObject@meta.data))
            }
            output$seurat <- DT::renderDataTable({
              df_show = NULL
              metadata_select = NULL
              req(input$readstats_select)
              rowCallback <- c(
                "function(row, data){",
                "  for(var i=0; i<data.length; i++){",
                "    if(data[i] === null){",
                "      $('td:eq('+i+')', row).html('NA')",
                "        .css({'color': 'rgb(151,151,151)', 'font-style': 'italic'});",
                "    }",
                "  }",
                "}"  
              )
              
              cols_not_needed = c('orig.ident')
              find_not_needed = which(colnames(metadata) %in% cols_not_needed)
              
              if(length(find_not_needed) > 0) {
                metadata = metadata %>% dplyr::select(!colnames(metadata)[find_not_needed])
              }

              if(input$readstats_select == "Read Stats") {
                cols = c('Barcoded_Reads', 'Trimmed_Reads', 'Unmapped_Reads', 'Mapped_Reads', 'Uniquely_Mapped_Reads', 'Multimapped_Reads', 'Chimeric_Reads', 'No_of_Reads_w_UMIs', 'Reads_After_Dedup_nUMIs_USSs', 'Ribosomal_Reads', 'Mitochondrial_Reads')

                find = which(colnames(metadata) %in% cols)
                if(length(find) > 0) {
  
                  metadata_select = metadata %>% dplyr::select(colnames(metadata)[find])
                  
                  columns_remove = c()
                  for(x in 1:length(colnames(metadata_select)))
                  {
                    number_of_NA = sum(is.na(metadata_select[[colnames(metadata_select)[x]]]))
                    if(number_of_NA == length(metadata_select[[colnames(metadata_select)[x]]])) {
                      getcol = colnames(metadata_select)[x]
                      columns_remove = c(columns_remove, getcol)
                    }
                  }
  
                  metadata_select = metadata_select %>% dplyr::select(-columns_remove)
                  table_Usability = data.frame(round(colSums(metadata_select)))
                  colnames(table_Usability) = "TotalCounts"
                  if('Ribosomal_Reads' %in% rownames(table_Usability)) {
                    table_Usability['Usable',] = table_Usability['Mapped_Reads',] - (table_Usability['Ribosomal_Reads',] + table_Usability['Mitochondrial_Reads',])
                    table_Usability['Undesirable',] = table_Usability['Ribosomal_Reads',] + table_Usability['Mitochondrial_Reads',]
                  } else {
                    table_Usability['Usable',] = table_Usability['Mapped_Reads',] - table_Usability['Mitochondrial_Reads',]
                    table_Usability['Undesirable',] = table_Usability['Mitochondrial_Reads',]
                  }
                  table_Usability$Percentage_Barcoded = round((table_Usability$TotalCounts/table_Usability['Barcoded_Reads',])*100, 2)
                  table_Usability$Percentage_Trimmed = round((table_Usability$TotalCounts/table_Usability['Trimmed_Reads',]$TotalCounts)*100, 2)
                  table_Usability['Barcoded_Reads',]$Percentage_Trimmed = 'NA'
                  table_Usability$TotalCounts = formatC(as.numeric(table_Usability$TotalCounts), big.mark = ",", big.interval = 3, format = "f", digits = 0)
                  colnames(table_Usability) = c('Total Counts', '% (of Barcoded Reads)', '% (of Trimmed Reads)')
                  df_show = table_Usability
                  colnames_use_datatable = colnames(table_Usability)
                } else {
                  metadata_select = NULL
                  df_show = NULL
                }
              }

              if(input$readstats_select == "Gene Body Assignment Breakdown") {
                cols = c('Mapped_Reads', 'Exon_nUMIs_USSs', 'Intron_nUMIs_USSs', 'Gene_nUMIs_USSs', 'Exon_Reads', 'Intron_Reads', 'Gene_Reads', 'Intergenic_Reads')

                find = which(colnames(metadata) %in% cols)
                if(length(find) > 0) {
  
                  metadata_select = metadata %>% dplyr::select(colnames(metadata)[find])
  
                  columns_remove = c()
                  for(x in 1:length(colnames(metadata_select)))
                  {
                    number_of_NA = sum(is.na(metadata_select[[colnames(metadata_select)[x]]]))
                    if(number_of_NA == length(metadata_select[[colnames(metadata_select)[x]]])) {
                      getcol = colnames(metadata_select)[x]
                      columns_remove = c(columns_remove, getcol)
                    }
                  }
  
                  metadata_select = metadata_select %>% dplyr::select(-columns_remove)
                  table_GB = data.frame(round(colSums(metadata_select)))
                  colnames(table_GB) = "TotalCounts"

                  table_GB$Percentage_Mapped = round((table_GB$TotalCounts/table_GB['Mapped_Reads',])*100, 2)
                  table_GB$TotalCounts = formatC(as.numeric(table_GB$TotalCounts), big.mark = ",", big.interval = 3, format = "f", digits = 0)
                  colnames(table_GB) = c('Total Counts', '% (of Mapped Reads)')
                  df_show = table_GB
                  colnames_use_datatable = colnames(table_GB)
                } else {
                  metadata_select = NULL
                  df_show = NULL
                }
                
              }

              if(input$readstats_select == "Undesirable Read Breakdown") {
                cols = c('Mapped_Reads', 'Ribosomal_Reads', 'Mitochondrial_Reads')

                find = which(colnames(metadata) %in% cols)
                if(length(find) > 0) {
  
                  metadata_select = metadata %>% dplyr::select(colnames(metadata)[find])
  
                  columns_remove = c()
                  for(x in 1:length(colnames(metadata_select)))
                  {
                    number_of_NA = sum(is.na(metadata_select[[colnames(metadata_select)[x]]]))
                    if(number_of_NA == length(metadata_select[[colnames(metadata_select)[x]]])) {
                      getcol = colnames(metadata_select)[x]
                      columns_remove = c(columns_remove, getcol)
                    }
                  }
  
                  metadata_select = metadata_select %>% dplyr::select(-columns_remove)
                  table_Undesirable = data.frame(round(colSums(metadata_select)))
                  colnames(table_Undesirable) = "TotalCounts"
                  if('Ribosomal_Reads' %in% rownames(table_Undesirable)) {
                    table_Undesirable['Usable',] = table_Undesirable['Mapped_Reads',] - (table_Undesirable['Ribosomal_Reads',] + table_Undesirable['Mitochondrial_Reads',])
                  } else {
                    table_Undesirable['Usable',] = table_Undesirable['Mapped_Reads',] - table_Undesirable['Mitochondrial_Reads',]
                  }
                  table_Undesirable$Percentage_Mapped = round((table_Undesirable$TotalCounts/table_Undesirable['Mapped_Reads',])*100, 2)
                  table_Undesirable$TotalCounts = formatC(as.numeric(table_Undesirable$TotalCounts), big.mark = ",", big.interval = 3, format = "f", digits = 0)
                  colnames(table_Undesirable) = c('Total Counts', '% (of Mapped Reads)')
                  df_show = table_Undesirable
                  colnames_use_datatable = colnames(table_Undesirable)
                } else {
                  metadata_select = NULL
                  df_show = NULL
                }
              }

              if(input$readstats_select == "Other Stats") {
                cols = c('No_of_Genes_nUMIs_USSs', 'No_of_Transcripts_nUMIs_USSs', 'No_of_Genes', 'No_of_Transcripts', 'Strand_Specificity')

                find = which(colnames(metadata) %in% cols)
                if(length(find) > 0) {
  
                  metadata_select = metadata %>% dplyr::select(colnames(metadata)[find])
  
                  columns_remove = c()
                  for(x in 1:length(colnames(metadata_select)))
                  {
                    number_of_NA = sum(is.na(metadata_select[[colnames(metadata_select)[x]]]))
                    if(number_of_NA == length(metadata_select[[colnames(metadata_select)[x]]])) {
                      getcol = colnames(metadata_select)[x]
                      columns_remove = c(columns_remove, getcol)
                    }
                  }
  
                  metadata_select = metadata_select %>% dplyr::select(-columns_remove)
                  table_other = data.frame(round(colMeans(metadata_select)))
                  colnames(table_other) = "AverageStats"
                  table_other$AverageStats = formatC(as.numeric(table_other$AverageStats), big.mark = ",", big.interval = 3, format = "f", digits = 0)
                  colnames(table_other) = 'Average Stats across barcodes'
                  df_show = table_other
                  colnames_use_datatable = colnames(table_other)
                } else {
                  metadata_select = NULL
                  df_show = NULL
                }
              }

              if(is.null(df_show) | is.null(metadata_select)) {
                shinyjs::disable("readstats_select")
                shinyalert("Uploded data is not processed using CogentAP v3.1", "Stats tables cannot be generated, downstream analysis will not be affected", type = "warning", confirmButtonCol = "#1c82e0")
                stop("Read Stats from CogentAP analysis not found; downstream analysis will not be impacted!")
              } else {
                df_show = data.frame(df_show)
                df_show[df_show == ""] <- NA
                df_show[df_show == "NaN"] <- NA
                df_show[df_show == 0] <- NA
                df_show = data.frame(df_show)
                DT::datatable(df_show, colnames = colnames_use_datatable, options = list(pageLength = 20, dom = "t", rowCallback = JS(rowCallback)))
              }
            })
            message("Seurat v5 object creation is completed!")
            data(seuratObject) # Update reactive value
          } else {
            message ("Fatal: Seurat object creation failed; Please check your inputs")
            stop("Error: Seurat object creation failed: Fatal")
          }
          incProgress(0.1, detail = 'checks complete, stats ready')
          timeEnd <- Sys.time()
          elapsedTime <- difftime(timeEnd, timeStart, units = 'mins')
        })
      }, error = function(e) {
        shinyjs::enable("upload")
        shinyalert("Data preparation failed", "Please check your inputs and parameters; Some possible errors: Metadata cells | data format | required input is empty", type = "error", confirmButtonCol = "#1c82e0")
      })
    })
    return(data)
  })
}
