
#' Dimension Reduction UI Component
#'
#' Generates UI elements for dimension reduction operations within a Shiny application.
#' This module allows users to perform Principal Component Analysis (PCA) and view
#' the elbow plot to determine the dimensionality of the dataset. Additional
#' plots for PCA loadings and a heatmap can be generated to further explore the
#' dimensionality reduced data.
#'
#' @param id A unique identifier for the UI module. This is necessary for distinguishing 
#'           this module's input from others in the same Shiny application.
#'
#' @return A UI layout comprising action buttons for PCA and elbow plot analysis,
#'         and output areas for displaying PCA, elbow plot, PCA loadings, 
#'         and heatmap visualizations.
#'         
#' @export
#' @examples
#' dimReductionUI("dimReductionModule")
dimReductionUI <- function(id) {
  ns <- NS(id)
  tagList(
    tabsetPanel(id = "PCAanalysis",
                tabPanel( "PCA",
                          h2("Linear Dimension Reduction"),
                          h5("*Required", style="color:#1c82e0"),
                          fluidRow(
                            column(
                              width = 4,
                              # Parameters for PCA
                              numericInput(ns("npcs"), "Number of principal components", value = 50),
                              bsPopover(ns("npcs"), "Number of principal components", "Total Number of PCs to compute and store", placement = "right", trigger = "hover", options = list(container = "body")),
                              checkboxInput(ns("approx"), "Approximate PCA", value = TRUE),
                              bsPopover(ns("approx"), "Approximation", "Use truncated singular value decomposition to approximate PCA", placement = "right", trigger = "hover", options = list(container = "body")),
                              selectInput(ns("PCAColorBy"), "Color by:", choices = NULL),
                              actionButton(ns("runPCAButton"), "Run PCA", class = "btn-primary", style='color:#fff; background-color:#1c82e0')
                            ),
                            column(
                              width = 8,
                              conditionalPanel(
                                condition = "input.runPCAButton > 0",
                                # PCA Plot output
                                selectInput(ns("pcaplot"), "File Type Selection:", choices = list("PNG" = ".png", "PDF" = ".pdf", "SVG" = ".svg", "JPEG" = ".jpeg"), selected = ".png"),
                                downloadButton(ns('downloadPlot_pcaplot'), 'Save plot', style='color:#fff; background-color:#1c82e0'),
                                plotlyOutput(ns("PCAPlotOutput"), width = 800, height = 600), ns = NS(id))))),
                tabPanel("Elbow Plot",
                         conditionalPanel(
                           condition = "input.runPCAButton > 0",
                           selectInput(ns("elbow"), "File Type Selection:", choices = list("PNG" = ".png", "PDF" = ".pdf", "SVG" = ".svg", "JPEG" = ".jpeg"), selected = ".png"),
                           downloadButton(ns('downloadPlot_elbow'), 'Save plot', style='color:#fff; background-color:#1c82e0'),
                           shiny::plotOutput(ns("elbowPlot"), width = 800, height = 600), ns = NS(id))),
                tabPanel("Viz Dim Loadings",
                         conditionalPanel(
                           condition = "input.runPCAButton > 0",
                           fluidRow(
                             column(
                               width = 3,
                               numericInput(ns("VizDims"), "Number of dimensions (PCs) to display", min = 1, max = 20, value = 2)),
                             column(width = 9,
                                    selectInput(ns("dimloadplot"), "File Type Selection:", choices = list("PNG" = ".png", "PDF" = ".pdf", "SVG" = ".svg", "JPEG" = ".jpeg"), selected = ".png"),
                                    downloadButton(ns('downloadPlot_dimloadplot'), 'Save plot', style='color:#fff; background-color:#1c82e0'),
                                    shiny::plotOutput(ns("dimLoadingsPlot")))), ns = NS(id))),
                tabPanel("PC Heatmap",
                         conditionalPanel(
                           condition = "input.runPCAButton > 0",
                           fluidRow(
                             column(width = 3,
                                    numericInput(ns("heatmapDims"), "Number of dimensions (PCs) to display", min = 1,
                                                 max = 10, value = 2),
                                    numericInput(ns("cells"), "Number of cells to consider", value = 500)),
                             column(width = 9,
                                    selectInput(ns("dimheatplot"), "File Type Selection:", choices = list("PNG" = ".png", "PDF" = ".pdf", "SVG" = ".svg", "JPEG" = ".jpeg"), selected = ".png"),
                                    downloadButton(ns('downloadPlot_dimheatplot'), 'Save plot', style='color:#fff; background-color:#1c82e0'),
                                    shiny::plotOutput(ns("dimHeatmapPlot")))), ns = NS(id)))))
}
