#' Gene fusion analysis module server logic
#' 
#' This server module handles overlaying fusion genes on UMAP
#'
#' @param id Unique identifier for the server module, ensuring isolated scope
#'           within the Shiny application.
#' @param seuratObj A reactive expression containing the Seurat object.
#'
#' @return
#' @export
#'
#' @examples
#' geneFusionServer("geneFusionModule", seuratObject)
geneFusionServer <- function(id, seuratObj) {
  moduleServer(id, function(input, output, session) {
    
    ns <- session$ns
    shinyjs::hide("FusionPlt")

    spanning_reads <- reactiveVal(NULL)
    junction_reads <- reactiveVal(NULL)
    
    output$downloadPlotUI <- renderUI({
      if (input$overlay_btn > 0 && !is.null(static_overlay_plot()) && !is.null(input$fusion_gene)) {
        downloadButton(ns('downloadPlot'), 'Save plot', style='color:#fff; background-color:#1c82e0')
      }
    })
    
    # Reactive value to hold metadata columns
    metadataColumns <- reactive({
      req(seuratObj())
      
      # Get the names of metadata columns
      all_metadata <- colnames(seuratObj()@meta.data)
      # Columns that are either factors or characters
      relevant_columns <- all_metadata[sapply(seuratObj()@meta.data[, all_metadata], function(col) is.factor(col) || is.character(col))]
      
      # Columns to be potentially removed
      column_not_needed <- c("Barcode")
      
      # Check and remove columns if they exist
      if (any(column_not_needed %in% relevant_columns)) {
        relevant_columns <- relevant_columns[!(relevant_columns %in% column_not_needed)]
      }
      
      # Return the filtered relevant columns
      relevant_columns
    })
    
    # Reactive expressions for spanning and junction reads
    spanning_reads <- reactive({
      req(seuratObj())
      if ("span_frags" %in% names(seuratObj()@assays)) {
        seuratObj()[["span_frags"]]$counts
      } else {
        NULL
      }
    })
    
    junction_reads <- reactive({
      req(seuratObj())
      if ("junction_reads" %in% names(seuratObj()@assays)) {
        seuratObj()[["junction_reads"]]$counts
      } else {
        NULL
      }
    })
    
    # Function to get fusion genes with additional information
    get_fusion_genes_with_info <- function() {
      req(seuratObj())
      withProgress(message = "Updating dropdown...", value = 0, {

        incProgress(0.6, detail = "Extracting data and running checks...")
        if (!is.null(spanning_reads()))  {
          span_counts <- colSums(spanning_reads())
          sorted_cells_span <- names(span_counts)[span_counts > 0]
        } else {
          sorted_cells_span = NULL
        }

        if (!is.null(junction_reads()))  {
          junction_counts <- colSums(junction_reads())
          sorted_cells_junction <- names(junction_counts)[junction_counts > 0]
        } else {
          sorted_cells_junction = NULL
        }

        if((length(sorted_cells_span) > 0) | (length(sorted_cells_junction) > 0)) {
          sorted_cells = unique(c(sorted_cells_span, sorted_cells_junction))
          fusion_genes_info = length(sorted_cells)
        } else {
          fusion_genes_info = NULL
        }
        incProgress(0.4, detail = "checks complete..")
        return(fusion_genes_info)
      })
    }
    
    fusion_genes_with_info <- reactive({
      get_fusion_genes_with_info()
    })
    
    # Dynamic populating of chocices
    observe({
      req(seuratObj(), input$overlay_type)  
      
      withProgress(message = "Updating dropdown...", value = 0, {
        # Extract spanning and junction read matrices
        incProgress(0.1, detail = "Extracting data...")
        
        # Ensure matrices are valid
        req(!is.null(spanning_reads()), !is.null(junction_reads()), 
            nrow(spanning_reads()) > 0, nrow(junction_reads()) > 0)
        
        # Calculate counts
        incProgress(0.1, detail = "Calculating span and junction counts...")
        span_counts <- rowSums(spanning_reads() > 0)
        junction_counts <- rowSums(junction_reads() > 0)
        
        # Filter genes based on the selected overlay type
        incProgress(0.1, detail = "Filtering and sorting genes...")
        filtered_genes <- switch(input$overlay_type,
                                 "Span Fragments" = {
                                   sorted_genes <- names(span_counts)[span_counts > 0]
                                   sorted_genes <- sorted_genes[order(-span_counts[sorted_genes])]
                                   if(length(sorted_genes) <= 5000) {
                                    sorted_genes
                                   } else {
                                    sorted_genes = sorted_genes[1:5000]
                                    sorted_genes
                                   }
                                 },
                                 "Junction Reads" = {
                                   sorted_genes <- names(junction_counts)[junction_counts > 0]
                                   sorted_genes <- sorted_genes[order(-junction_counts[sorted_genes])]
                                   if(length(sorted_genes) <= 5000) {
                                    sorted_genes
                                   } else {
                                    sorted_genes = sorted_genes[1:5000]
                                    sorted_genes
                                   }
                                 },
                                 "Span Fragments + Junction Reads" = {
                                   sorted_genes_span <- names(span_counts)[span_counts > 0]
                                   sorted_genes_junction <- names(junction_counts)[junction_counts > 0]
                                   sorted_genes_temp <- intersect(sorted_genes_span, sorted_genes_junction)
                                   sorted_genes_temp <- sorted_genes_temp[order(-junction_counts[sorted_genes_temp], -span_counts[sorted_genes_temp])]
                                   sorted_genes = c()
                                   for (gene in sorted_genes_temp) {
                                    cells_with_fusion_check_span <- names(which(spanning_reads()[gene, ] > 0))
                                    cells_with_fusion_check_junction <- names(which(junction_reads()[gene, ] > 0))
                                    cells_with_fusion_check = intersect(cells_with_fusion_check_span, cells_with_fusion_check_junction)
                                    if(length(cells_with_fusion_check) > 0) {
                                      sorted_genes = c(sorted_genes, gene)
                                    }
                                   }
                                   if(length(sorted_genes) <= 5000) {
                                    sorted_genes
                                   } else {
                                    sorted_genes = sorted_genes[1:5000]
                                    sorted_genes
                                   }
                                 })
        
        # Handling of proper displaying of message "No valid genes"
        updateSelectizeInput(session, "fusion_gene", 
                             choices = NULL, options = list(placeholder = ""),  
                             server = T)
        
        # Handle case when no valid genes are available
        if (length(filtered_genes) == 0) {
          shinyjs::disable("overlay_btn")  # Disable overlay button
          updateSelectizeInput(session, "fusion_gene", choices = NULL, options = list(placeholder = "No valid genes available"), server = T)
          return()  
        }
        
        # Update dropdown and enable button when valid genes are available
        shinyjs::enable("overlay_btn")  # Enable overlay button
        incProgress(0.1, detail = "Creating dropdown options...")
        choices <- lapply(filtered_genes, function(gene) {
          paste(gene, "(Span:", span_counts[gene], ", Junction:", junction_counts[gene], ")")
        })
        updateSelectizeInput(session, "fusion_gene", choices = choices, server = TRUE)
        incProgress(1, detail = "Finalizing updates...")
      })
    })
    
    
    output$ConditionalParamteres <- renderUI({
      if (is.null(fusion_genes_with_info()) || length(fusion_genes_with_info()) == 0) {
        tagList(
          HTML("<p>No fusion gene data available</p>")
        )
      } else {
        column(4,
               selectizeInput(ns("fusion_gene"), label = "Select fusion genes:",
                           choices = NULL, multiple = TRUE),
               radioButtons(ns("overlay_type"), "Overlay type:",
                            choices = c("Span Fragments", "Junction Reads", "Span Fragments + Junction Reads"),
                            selected = "Span Fragments"), 
               selectInput(ns("umapColorBy"), "Color by:", choices = metadataColumns()),
               actionButton(ns("overlay_btn"), "Overlay Plot", style='color:#fff; background-color:#1c82e0')
        )
      }
    })
    
    
    # Reactive expression to subset umap coordinates with cells having the selected fusion genes
    umap_coords_with_fusion <- reactive({
      req(input$fusion_gene, seuratObj())
      
      # Initialize an empty data frame to store the coordinates for valid cells
      umap_coords_with_fusion <- data.frame(matrix(ncol = 2, nrow = 0))
      
      selected_genes <- gsub("\\s*\\(.*", "", input$fusion_gene)  # Clean up the gene names
      
      # Loop over each gene and process the corresponding cells
      for (gene in selected_genes) {
        
        # Initialize variable for the cells to be returned
        cells_with_fusion <- NULL
        
        # Reset cells based on selected overlay type
        if (input$overlay_type == "Span Fragments") {
          # Cells with spanning reads only
          cells_with_fusion <- which(spanning_reads()[gene, ] > 0)
        } else if (input$overlay_type == "Junction Reads") {
          # Cells with junction reads only
          cells_with_fusion <- which(junction_reads()[gene, ] > 0)
        } else if (input$overlay_type == "Span Fragments + Junction Reads") {
          # Cells with both spanning and junction reads
          combined_cells <- which(spanning_reads()[gene, ] > 0 & junction_reads()[gene, ] > 0)
          
          if (length(combined_cells) == 0) {
            # shinyalert("No Cells Found",
            #            "No cells found with both spanning and junction reads",
            #            type = "warning",
            #            confirmButtonCol = "#1c82e0"
            # )
            cells_with_fusion <- NULL
          } else {
            cells_with_fusion <- combined_cells
          }
        }
        
        # Skip gene if no cells were selected
        if (length(cells_with_fusion) == 0) {
          next
        }
        
        # Add UMAP coordinates for the selected cells
        if("umap" %in% Reductions(seuratObj())) {
          umap_coords <- data.frame(seuratObj()[["umap"]]@cell.embeddings)
          umap_coords_with_fusion_gene <- data.frame(umap_coords[rownames(umap_coords) %in% colnames(spanning_reads())[cells_with_fusion], ])
          
          # Ensure valid data before adding
          if (nrow(umap_coords_with_fusion_gene) > 0) {
            umap_coords_with_fusion_gene$gene <- gene
            umap_coords_with_fusion_gene$gene <- as.factor(umap_coords_with_fusion_gene$gene)
            umap_coords_with_fusion <- rbind(umap_coords_with_fusion, umap_coords_with_fusion_gene)
          }
        }
        if("tsne" %in% Reductions(seuratObj())) {
          umap_coords <- data.frame(seuratObj()[["tsne"]]@cell.embeddings)
          umap_coords_with_fusion_gene <- data.frame(umap_coords[rownames(umap_coords) %in% colnames(spanning_reads())[cells_with_fusion], ])
          
          # Ensure valid data before adding
          if (nrow(umap_coords_with_fusion_gene) > 0) {
            umap_coords_with_fusion_gene$gene <- gene
            umap_coords_with_fusion_gene$gene <- as.factor(umap_coords_with_fusion_gene$gene)
            umap_coords_with_fusion <- rbind(umap_coords_with_fusion, umap_coords_with_fusion_gene)
          }
        }
      }
      
      return(umap_coords_with_fusion)
    })
    
    # Reactive expression for overlay plot
    overlay_plot <- eventReactive(input$overlay_btn, {
      
      if (input$overlay_btn > 0) {
        if (is.null(input$fusion_gene)) {
          shinyalert("Error", "Please select fusion genes to overlay.", type = "error", confirmButtonCol = "#1c82e0")
          return()
        }
      }
      req(umap_coords_with_fusion(), seuratObj(), input$fusion_gene, input$umapColorBy)
      
      # Check if either UMAP or tSNE reductions are present
      if (!"umap" %in% Reductions(seuratObj()) && !"tsne" %in% Reductions(seuratObj())) {
        shinyalert("Neither UMAP nor t-SNE reduction is available. Please run clustering Module.", type = "warning", confirmButtonCol = "#1c82e0")
        return()
      }
      
      # Create the UMAP plot
      p1 <- DimPlot(seuratObj(), pt.size = 1.5, label = FALSE, label.size = 5 ,group.by = input$umapColorBy) + 
        RotatedAxis() + 
        guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) + 
        theme(aspect.ratio = 1, axis.text = element_text(colour = "black", size = 15, face = "bold"), 
              axis.title = element_text(colour = "black", size = 15, face = "bold"), 
              legend.position = "right", legend.box = "vertical", 
              panel.background = element_rect(colour = "white"))
      
      # Convert ggplot to plotly
      plot <- ggplotly(p1)
      
      selected_genes <- gsub("\\s*\\(.*", "", input$fusion_gene)
      for (gene in selected_genes) {
        if("umap" %in% Reductions(seuratObj())) {
          plot <- plot %>% add_trace(
            data = umap_coords_with_fusion()[umap_coords_with_fusion()$gene == gene, ],
            x = ~umap_1,
            y = ~umap_2,
            type = "scatter",
            mode = "markers",
            marker = list(
              size = 7,
              symbol = "circle-open",
              color = "black"
            ),
            hoverinfo = "text",
            text = ~gene,
            name = gene
          )
        }
        if("tsne" %in% Reductions(seuratObj())) {
          plot <- plot %>% add_trace(
            data = umap_coords_with_fusion()[umap_coords_with_fusion()$gene == gene, ],
            x = ~tSNE_1,
            y = ~tSNE_2,
            type = "scatter",
            mode = "markers",
            marker = list(
              size = 7,
              symbol = "circle-open",
              color = "black"
            ),
            hoverinfo = "text",
            text = ~gene,
            name = gene
          )
        }
      }
      plot
    })
    
    # Reactive expression to generate the static overlay plot
    static_overlay_plot <- reactive({
      req(umap_coords_with_fusion(), seuratObj(), input$fusion_gene, input$umapColorBy)
      
      p1 <- DimPlot(seuratObj(), pt.size = 1.5, label = FALSE, label.size = 5, group.by = input$umapColorBy) + 
        RotatedAxis() + guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) + 
        theme(aspect.ratio = 1, axis.text = element_text(colour = "black", size = 15, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white"))
      
      selected_genes <- gsub("\\s*\\(.*", "", input$fusion_gene)
      
      for (gene in selected_genes) {
        if("umap" %in% Reductions(seuratObj())) {
          p1 <- p1 + geom_point(
            data = umap_coords_with_fusion()[umap_coords_with_fusion()$gene == gene, ],
            aes(x = umap_1, y = umap_2, shape = gene),
            size = 2,
            fill = NA,
            color = "black") + scale_shape_manual(values = rep(21, length(input$fusion_gene))) +
            guides(color = guide_legend(order=1),
                   shape = guide_legend(order=2, title = "")) +
            theme(legend.spacing.y = unit(-0.2, "cm"))
          
        }
        if("tsne" %in% Reductions(seuratObj())) {
          p1 <- p1 + geom_point(
            data = umap_coords_with_fusion()[umap_coords_with_fusion()$gene == gene, ],
            aes(x = tSNE_1, y = tSNE_2, shape = gene),
            size = 2,
            fill = NA,
            color = "black") + scale_shape_manual(values = rep(21, length(input$fusion_gene))) +
            guides(shape = guide_legend(order=2, title = "")) +
            theme(legend.spacing.y = unit(-0.2, "cm"))
          
        }
      }
      p1
    })
    
    # Render fusion gene plot
    output$fusion_plot <- renderPlotly({
      overlay_plot()
    })
    
    observe({
      if (!is.null(overlay_plot())) {
        shinyjs::show("FusionPlt")
      }
    })
    
    output$downloadPlot <- downloadHandler(
      filename = function() {
        paste0("CogentDS_FusionOverlayPlot", "-", Sys.Date(), input$FusionPlt)
      },
      content = function(file) {
        ggsave(file, plot = static_overlay_plot(), width = 10, height = 10)
      },
      contentType = 'image'
    )
  })
}